# Copyright 2019 Xanadu Quantum Technologies Inc.

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
Hermite Multidimensional Python interface
"""
from typing import Tuple, Generator, Iterable
from numba import jit
from numba.cpython.unsafe.tuple import tuple_setitem
import numpy as np

from ._hafnian import input_validation


# pylint: disable=too-many-arguments
def hermite_multidimensional(
    R, cutoff, y=None, C=1, renorm=False, make_tensor=True, modified=False, rtol=1e-05, atol=1e-08
):
    r"""Returns photon number statistics of a Gaussian state for a given
    covariance matrix as described in *Multidimensional Hermite polynomials
    and photon distribution for polymode mixed light*
    `arxiv:9308033 <https://arxiv.org/abs/hep-th/9308033>`_.

    Here :math:`R` is an :math:`n \times n` square matrix, and
    :math:`y` is an :math:`n` dimensional vector. The polynomials :math:`H_k^{(R)}(y)` are
    parametrized by the multi-index :math:`k=(k_0,k_1,\ldots,k_{n-1})`,
    and are calculated for all values :math:`0 \leq k_j < \text{cutoff}`,
    thus a tensor of dimensions :math:`\text{cutoff}^n` is returned.

    This tensor can either be flattened into a vector or returned as an actual
    tensor with :math:`n` indices.

    This implementation is based on the MATLAB code available at github
    `clementsw/gaussian-optics <https://github.com/clementsw/gaussian-optics>`_.

    .. note::

        Note that if :math:`R = (1)` then :math:`H_k^{(R)}(y)`
        are precisely the well known **probabilists' Hermite polynomials** :math:`He_k(y)`,
        and if :math:`R = (2)` then :math:`H_k^{(R)}(y)` are precisely the well known
        **physicists' Hermite polynomials** :math:`H_k(y)`.

    Args:
        R (array): square matrix parametrizing the Hermite polynomial family
        cutoff (int): maximum size of the subindices in the Hermite polynomial
        y (array): vector argument of the Hermite polynomial
        C (complex): first value of the Hermite polynomials, the default value is 1
        renorm (bool): If ``True``, normalizes the returned multidimensional Hermite
            polynomials such that :math:`H_k^{(R)}(y)/\prod_i k_i!`
        make_tensor (bool): If ``False``, returns a flattened one dimensional array
            containing the values of the polynomial
        modified (bool): whether to return the modified multidimensional Hermite polynomials or the standard ones
        rtol (float): the relative tolerance parameter used in ``np.allclose``
        atol (float): the absolute tolerance parameter used in ``np.allclose``
    Returns:
        (array): the multidimensional Hermite polynomials
    """

    input_validation(R, atol=atol, rtol=rtol)
    n, _ = R.shape

    if (modified is False) and (y is not None):
        m = y.shape[0]
        if m == n:
            ym = R @ y
            return hermite_multidimensional(
                R, cutoff, y=ym, C=C, renorm=renorm, make_tensor=make_tensor, modified=True
            )

    if y is None:
        y = np.zeros([n], dtype=complex)

    m = y.shape[0]
    if m != n:
        raise ValueError("The matrix R and vector y have incompatible dimensions")

    num_indices = len(y)
    # we want to catch np.ndarray(int) of ndim=0 which cannot be cast to tuple
    if isinstance(cutoff, np.ndarray) and (cutoff.ndim == 0 or len(cutoff) == 1):
        cutoff = int(cutoff)
    if isinstance(cutoff, Iterable):
        cutoffs = tuple(cutoff)
    else:
        cutoffs = tuple([cutoff]) * num_indices

    Rt = np.real_if_close(R)
    yt = np.real_if_close(y)

    dtype = np.find_common_type([Rt.dtype.name, yt.dtype.name], [np.array(C).dtype.name])
    array = np.zeros(cutoffs, dtype=dtype)
    array[(0,) * num_indices] = C

    if renorm:
        values = np.array(_hermite_multidimensional_renorm(Rt, yt, array))
    else:
        values = np.array(_hermite_multidimensional(Rt, yt, array))

    if not make_tensor:
        values = values.flatten()

    return values


def interferometer(R, cutoff, C=1, renorm=True, make_tensor=True, rtol=1e-05, atol=1e-08):
    r"""Returns the matrix elements of an interferometer parametrized in terms of its R matrix.

    Here :math:`R` is an :math:`n \times n` square matrix. The polynomials are
    parametrized by the multi-index :math:`k=(k_0,k_1,\ldots,k_{n-1})`,
    and are calculated for all values :math:`0 \leq k_j < \text{cutoff}`,
    thus a tensor of dimensions :math:`\text{cutoff}^n` is returned.

    This tensor can either be flattened into a vector or returned as an actual
    tensor with :math:`n` indices.

    .. note::

        Note that `interferometer` uses the normalized multidimensional Hermite polynomials.

    Args:
        R (array): square matrix parametrizing the Hermite polynomial family
        cutoff (int): maximum size of the subindices in the Hermite polynomial
        C (complex): first value of the Hermite polynomials, the default value is 1
        renorm (bool): If ``True``, normalizes the returned multidimensional Hermite
            polynomials such that :math:`H_k^{(R)}(y)/\prod_i k_i!`
        make_tensor (bool): If ``False``, returns a flattened one dimensional array
            containing the values of the polynomial
        rtol (float): the relative tolerance parameter used in ``np.allclose``
        atol (float): the absolute tolerance parameter used in ``np.allclose``
    Returns:
        (array): the multidimensional Hermite polynomials
    """

    input_validation(R, atol=atol, rtol=rtol)
    n, num_indices = R.shape

    # we want to catch np.ndarray(int) of ndim=0 which cannot be cast to tuple
    if isinstance(cutoff, np.ndarray) and (cutoff.ndim == 0 or len(cutoff) == 1):
        cutoff = int(cutoff)
    if isinstance(cutoff, Iterable):
        cutoffs = tuple(cutoff)
    else:
        cutoffs = tuple([cutoff]) * num_indices

    Rt = np.real_if_close(R)

    dtype = np.find_common_type([Rt.dtype.name], [np.array(C).dtype.name])
    array = np.zeros(cutoffs, dtype=dtype)
    array[(0,) * num_indices] = C

    if renorm:
        values = np.array(_interferometer_renorm(Rt, array))
    else:
        values = np.array(_interferometer(Rt, array))

    if make_tensor:
        shape = cutoff * np.ones([n], dtype=int)
        values = np.reshape(values, shape)

    return values


# pylint: disable=too-many-arguments
def hafnian_batched(A, cutoff, mu=None, rtol=1e-05, atol=1e-08, renorm=False, make_tensor=True):
    r"""Calculates the hafnian of :func:`reduction(A, k) <hafnian.reduction>`
    for all possible values of vector ``k`` below the specified cutoff.

    Here,

    * :math:`A` is am :math:`n\times n` square matrix
    * :math:`k` is a vector of (non-negative) integers with the same dimensions as :math:`A`,
      i.e., :math:`k = (k_0,k_1,\ldots,k_{n-1})`, and where :math:`0 \leq k_j < \texttt{cutoff}`.

    The function :func:`~.hafnian_repeated` can be used to calculate the reduced hafnian
    for a *specific* value of :math:`k`; see the documentation for more information.

    .. note::

        If ``mu`` is not ``None``, this function instead returns
        ``hafnian(np.fill_diagonal(reduction(A, k), reduction(mu, k)), loop=True)``.
        This calculation can only be performed if the matrix :math:`A` is invertible.

    Args:
        A (array): a square, symmetric :math:`N\times N` array.
        cutoff (int): maximum size of the subindices in the Hermite polynomial
        mu (array): a vector of length :math:`N` representing the vector of means/displacement
        renorm (bool): If ``True``, the returned hafnians are *normalized*, that is,
            :math:`haf(reduction(A, k))/\ \sqrt{prod_i k_i!}`
            (or :math:`lhaf(fill\_diagonal(reduction(A, k),reduction(mu, k)))` if
            ``mu`` is not None)
        make_tensor: If ``False``, returns a flattened one dimensional array instead
            of a tensor with the values of the hafnians.
        rtol (float): the relative tolerance parameter used in ``np.allclose``.
        atol (float): the absolute tolerance parameter used in ``np.allclose``.
    Returns:
        (array): the values of the hafnians for each value of :math:`k` up to the cutoff
    """
    input_validation(A, atol=atol, rtol=rtol)
    n, _ = A.shape

    if mu is None:
        mu = np.zeros([n], dtype=complex)
    # The minus signs are intentional and are due to the fact that
    # Dodonov et al. in PRA 50, 813 (1994) use (p,q) ordering instead of (q,p) ordering
    return hermite_multidimensional(
        -A, cutoff, y=mu, renorm=renorm, make_tensor=make_tensor, modified=True
    )


@jit(nopython=True)
def dec(tup: Tuple[int], i: int) -> Tuple[int, ...]:  # pragma: no cover
    r"""returns a copy of the given tuple of integers where the ith element has been decreased by 1
    Args:
        tup (Tuple[int]): the given tuple
        i (int): the position of the element to be decreased
    Returns:
        Tuple[int,...]: the new tuple with the decrease on i-th element by 1
    """
    copy = tup[:]
    return tuple_setitem(copy, i, tup[i] - 1)


@jit(nopython=True)
def remove(
    pattern: Tuple[int, ...]
) -> Generator[Tuple[int, Tuple[int, ...]], None, None]:  # pragma: no cover
    r"""returns a generator for all the possible ways to decrease elements of the given tuple by 1
    Args:
        pattern (Tuple[int, ...]): the pattern given to be decreased
    Returns:
        Generator[Tuple[int, Tuple[int, ...]], None, None]: the generator
    """
    for p, n in enumerate(pattern):
        if n > 0:
            yield p, dec(pattern, p)


SQRT = np.sqrt(np.arange(1000))  # saving the time to recompute square roots


@jit(nopython=True)
def _hermite_multidimensional_renorm(R, y, G):  # pragma: no cover
    r"""Numba-compiled function to fill an array with the Hermite polynomials. It expects an array
    initialized with zeros everywhere except at index (0,...,0) (i.e. the seed value).

    Args:
        R (array[complex]): square matrix parametrizing the Hermite polynomial
        y (vector[complex]): vector argument of the Hermite polynomial
        G (array[complex]): array to be filled with the Hermite polynomials

    Returns:
        array[complex]: the multidimensional Hermite polynomials
    """
    indices = np.ndindex(G.shape)
    next(indices)  # skip the first index (0,...,0)
    for idx in indices:
        i = 0
        for i, val in enumerate(idx):
            if val > 0:
                break
        ki = dec(idx, i)
        u = y[i] * G[ki]
        for l, kl in remove(ki):
            u -= SQRT[ki[l]] * R[i, l] * G[kl]
        G[idx] = u / SQRT[idx[i]]
    return G


@jit(nopython=True)
def _hermite_multidimensional(R, y, G):  # pragma: no cover
    r"""Numba-compiled function to fill an array with the Hermite polynomials. It expects an array
    initialized with zeros everywhere except at index (0,...,0) (i.e. the seed value).

    Args:
        R (array[complex]): square matrix parametrizing the Hermite polynomial
        y (vector[complex]): vector argument of the Hermite polynomial
        G (array[complex]): array to be filled with the Hermite polynomials

    Returns:
        array[complex]: the multidimensional Hermite polynomials
    """
    indices = np.ndindex(G.shape)
    next(indices)  # skip the first index (0,...,0)
    for idx in indices:
        i = 0
        for i, val in enumerate(idx):
            if val > 0:
                break
        ki = dec(idx, i)
        u = y[i] * G[ki]
        for l, kl in remove(ki):
            u -= ki[l] * R[i, l] * G[kl]
        G[idx] = u
    return G


@jit(nopython=True)
def _interferometer_renorm(R, G):  # pragma: no cover
    r"""Numba-compiled function returning the matrix elements of an interferometer
    parametrized in terms of its R matrix

    Args:
        R (array[complex]): square matrix parametrizing the Hermite polynomial
        array (array[complex]): array to be filled with the Hermite polynomials

    Returns:
        array[complex]: the multidimensional Hermite polynomials
    """
    dim, _ = R.shape
    num_modes = dim / 2

    indices = np.ndindex(G.shape)
    next(indices)  # skip the first index (0,...,0)
    for idx in indices:
        bran = 0
        for ii in range(0, num_modes):
            bran += idx[ii]

        ketn = 0
        for ii in range(num_modes, dim):
            ketn += idx[ii]

        if bran == ketn:
            i = 0
            for i, val in enumerate(idx):
                if val > 0:
                    break
            ki = dec(idx, i)
            u = 0
            for l, kl in remove(ki):
                u -= SQRT[ki[l]] * R[i, l] * G[kl]
            G[idx] = u / SQRT[idx[i]]

    return G


@jit(nopython=True)
def _interferometer(R, G):  # pragma: no cover
    r"""Numba-compiled function returning the matrix elements of an interferometer
    parametrized in terms of its R matrix

    Args:
        R (array[complex]): square matrix parametrizing the Hermite polynomial
        G (array[complex]): array to be filled with the Hermite polynomials

    Returns:
        array[complex]: the multidimensional Hermite polynomials
    """
    dim, _ = R.shape
    num_modes = dim / 2

    indices = np.ndindex(G.shape)
    next(indices)  # skip the first index (0,...,0)
    for idx in indices:
        bran = 0
        for ii in range(0, num_modes):
            bran += idx[ii]

        ketn = 0
        for ii in range(num_modes, dim):
            ketn += idx[ii]

        if bran == ketn:
            i = 0
            for i, val in enumerate(idx):
                if val > 0:
                    break
            ki = dec(idx, i)
            u = 0
            for l, kl in remove(ki):
                u -= ki[l] * R[i, l] * G[kl]
            G[idx] = u

    return G


def grad_hermite_multidimensional(G, R, y, C=1, renorm=True, dtype=None):
    # pylint: disable=too-many-arguments
    r"""Calculates the gradients of the renormalized multidimensional Hermite polynomials :math:`C*H_k^{(R)}(y)` with respect to its parameters :math:`C`, :math:`y` and :math:`R`.

    Args:
        G (array): the multidimensional Hermite polynomials
        R (array[complex]): square matrix parametrizing the Hermite polynomial
        y (vector[complex]): vector argument of the Hermite polynomial
        C (complex): first value of the Hermite polynomials
        renorm (bool): If ``True``, uses the normalized multidimensional Hermite
            polynomials such that :math:`H_k^{(R)}(y)/\prod_i k_i!`
        dtype (data type): Specifies the data type used for the calculation

    Returns:
        array[data type], array[data type], array[data type]: the gradients of the multidimensional Hermite polynomials with respect to C, R and y
    """
    if dtype is None:
        dtype = np.find_common_type(
            [G.dtype.name, R.dtype.name, y.dtype.name], [np.array(C).dtype.name]
        )
    n, _ = R.shape
    if y.shape[0] != n:
        raise ValueError(
            f"The matrix R and vector y have incompatible dimensions ({R.shape} vs {y.shape})"
        )
    dG_dC = np.array(G / C).astype(dtype)
    dG_dR = np.zeros(G.shape + R.shape, dtype=dtype)
    dG_dy = np.zeros(G.shape + y.shape, dtype=dtype)
    if renorm:
        dG_dR, dG_dy = _grad_hermite_multidimensional_renorm(R, y, G, dG_dR, dG_dy)
    else:
        dG_dR, dG_dy = _grad_hermite_multidimensional(R, y, G, dG_dR, dG_dy)

    return dG_dC, dG_dR, dG_dy


@jit(nopython=True)
def _grad_hermite_multidimensional_renorm(R, y, G, dG_dR, dG_dy):  # pragma: no cover
    r"""
    Numba-compiled function to fill two arrays (dG_dR, dG_dy) with the gradients of the renormalized multidimensional Hermite polynomials
    with respect to its parameters :math:`R` and :math:`y`. It needs the `array` of the multidimensional Hermite polynomials.

    Args:
        R (array[complex]): square matrix parametrizing the Hermite polynomial
        y (vector[complex]): vector argument of the Hermite polynomial
        G (array[complex]): array of the multidimensional Hermite polynomials
        dG_dR (array[complex]): array to be filled with the gradients of the renormalized multidimensional Hermite polynomials with respect to R
        dG_dy (array[complex]): array to be filled with the gradients of the renormalized multidimensional Hermite polynomials with respect to y

    Returns:
        dG_dR[complex], dG_dy[complex]: the gradients of the renormalized multidimensional Hermite polynomials with respect to R and y
    """
    indices = np.ndindex(G.shape)
    next(indices)  # skip the first index (0,...,0)
    for idx in indices:
        i = 0
        for i, val in enumerate(idx):
            if val > 0:
                break
        ki = dec(idx, i)
        dy = y[i] * dG_dy[ki]
        dy[i] += G[ki]
        dR = y[i] * dG_dR[ki]
        for l, kl in remove(ki):
            dy -= SQRT[ki[l]] * dG_dy[kl] * R[i, l]
            dR -= SQRT[ki[l]] * R[i, l] * dG_dR[kl]
            dR[i, l] -= SQRT[ki[l]] * G[kl]
        dG_dR[idx] = dR / SQRT[idx[i]]
        dG_dy[idx] = dy / SQRT[idx[i]]
    return dG_dR, dG_dy


@jit(nopython=True)
def _grad_hermite_multidimensional(R, y, G, dG_dR, dG_dy):  # pragma: no cover
    r"""
    Numba-compiled function to fill two arrays (dG_dR, dG_dy) with the gradients of the renormalized multidimensional Hermite polynomials
    with respect to its parameters :math:`R` and :math:`y`. It needs the `array` of the multidimensional Hermite polynomials.

    Args:
        R (array[complex]): square matrix parametrizing the Hermite polynomial
        y (vector[complex]): vector argument of the Hermite polynomial
        G (array[complex]): array of the multidimensional Hermite polynomials
        dG_dR (array[complex]): array to be filled with the gradients of the renormalized multidimensional Hermite polynomials with respect to R
        dG_dy (array[complex]): array to be filled with the gradients of the renormalized multidimensional Hermite polynomials with respect to y

    Returns:
        dG_dR[complex], dG_dy[complex]: the gradients of the renormalized multidimensional Hermite polynomials with respect to R and y
    """
    indices = np.ndindex(G.shape)
    next(indices)  # skip the first index (0,...,0)
    for idx in indices:
        i = 0
        for i, val in enumerate(idx):
            if val > 0:
                break
        ki = dec(idx, i)
        dy = y[i] * dG_dy[ki]
        dy[i] += G[ki]
        dR = y[i] * dG_dR[ki]
        for l, kl in remove(ki):
            dy -= ki[l] * dG_dy[kl] * R[i, l]
            dR -= ki[l] * R[i, l] * dG_dR[kl]
            dR[i, l] -= ki[l] * G[kl]
        dG_dR[idx] = dR
        dG_dy[idx] = dy
    return dG_dR, dG_dy



# Copyright 2019-2020 Xanadu Quantum Technologies Inc.

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
Functions for rescaling adjacency matrices as well as calculating the Qmat
covariance matrix from an adjacency matrix.
"""

import numpy as np
from scipy.optimize import root_scalar

from .conversions import Xmat, Covmat
from .means_and_variances import mean_clicks


def adj_scaling_torontonian(A, c_mean):
    r"""Returns the scaling parameter by which the adjacency matrix A
    should be rescaled so that the Gaussian state that encodes it has
    give a mean number of clicks equal to ``c_mean`` when measured with
    threshold detectors.

    Args:
        A (array): adjacency matrix
        c_mean (float): mean photon number of the Gaussian state

    Returns:
        float: scaling parameter
    """
    n, _ = A.shape
    if c_mean < 0 or c_mean > n:
        raise ValueError("The mean number of clicks should be smaller than the number of modes")

    vals = np.linalg.svd(A, compute_uv=False)
    localA = A / vals[0]  # rescale the matrix so that the singular values are between 0 and 1.

    def cost(x):
        r"""Cost function giving the difference between the wanted number of clicks and the number
        of clicks at a given scaling value. It assumes that the adjacency matrix has been rescaled
        so that it has singular values between 0 and 1.

        Args:
            x (float): scaling value

        Return:
            float: difference between desired and obtained mean number of clicks
        """
        if x >= 1.0:
            return c_mean - n
        if x <= 0:
            return c_mean
        return c_mean - _mean_clicks_adj(x * localA)

    res = root_scalar(cost, x0=0.5, bracket=(0.0, 1.0))  # Do the optimization

    if not res.converged:
        raise ValueError("The search for a scaling value failed")
    return res.root / vals[0]


def _mean_clicks_adj(A):
    r"""Given an adjacency matrix this function calculates the mean number of clicks.
    For this to make sense the user must provide a matrix with singular values
    less than or equal to one. See Appendix A.3 of <https://arxiv.org/abs/1902.00462>`_
    by Banchi et al.

    Args:
        A (array): rescaled adjacency matrix

    Returns:
        float: mean number of clicks
    """
    n, _ = A.shape
    idn = np.identity(n)
    X = np.block([[0 * idn, idn], [idn, 0 * idn]])
    B = np.block([[A, 0 * A], [0 * A, np.conj(A)]])
    Q = np.linalg.inv(np.identity(2 * n) - X @ B)
    return mean_clicks(Covmat(Q))


def adj_scaling(A, n_mean):
    r"""Returns the scaling parameter by which the adjacency matrix A
    should be rescaled so that the Gaussian state that endodes it has
    a total mean photon number n_mean.

    Args:
        A (array): Adjacency matrix
        n_mean (float): Mean photon number of the Gaussian state

    Returns:
        float: Scaling parameter
    """
    eps = 1e-10
    ls = np.linalg.svd(A, compute_uv=False)
    max_sv = ls[0]
    a_lim = 0.0
    b_lim = 1.0 / (eps + max_sv)
    x_init = 0.5 * b_lim

    if 1000 * eps >= max_sv:
        raise ValueError("The singular values of the matrix A are too small.")

    def mean_photon_number(x, vals):
        r"""Returns the mean number of photons in the Gaussian state that
        encodes the adjacency matrix x*A where vals are the singular values of A

        Args:
            x (float): Scaling parameter
            vals (array): Singular values of the matrix A

        Returns:
            n_mean: Mean photon number in the Gaussian state
        """
        vals2 = (x * vals) ** 2
        n = np.sum(vals2 / (1.0 - vals2))
        return n

    # The following function is implicitly tested in test_adj_scaling
    def grad_mean_photon_number(x, vals):  # pragma: no cover
        r"""Returns the gradient od the mean number of photons in the Gaussian state that
        encodes the adjacency matrix x*A with respect to x.
        vals are the singular values of A

        Args:
            x (float): Scaling parameter
            vals (array): Singular values of the matrix A

        Returns:
            d_n_mean: Derivative of the mean photon number in the Gaussian state
                with respect to x
        """
        vals1 = vals * x
        dn = (2.0 / x) * np.sum((vals1 / (1 - vals1 ** 2)) ** 2)
        return dn

    f = lambda x: mean_photon_number(x, ls) - n_mean
    df = lambda x: grad_mean_photon_number(x, ls)
    res = root_scalar(f, fprime=df, x0=x_init, bracket=(a_lim, b_lim))

    if not res.converged:
        raise ValueError("The search for a scaling value failed")

    return res.root


def adj_to_qmat(A, n_mean):
    r"""Returns the Qmat xp-covariance matrix associated to a graph with
    adjacency matrix :math:`A` and with mean photon number :math:`n_{mean}`.

    Args:
        A (array): a :math:`N\times N` ``np.float64`` (symmetric) adjacency matrix
        n_mean (float): mean photon number of the Gaussian state

    Returns:
        array: the :math:`2N\times 2N` Q matrix.
    """
    n, m = A.shape

    if n != m:
        raise ValueError("Matrix must be square.")

    sc = adj_scaling(A, n_mean)
    Asc = sc * A
    A = np.block([[Asc, 0 * Asc], [0 * Asc, Asc.conj()]])
    I = np.identity(2 * n)
    X = Xmat(n)
    Q = np.linalg.inv(I - X @ A)
    return Q



# Copyright 2019 Xanadu Quantum Technologies Inc.

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
Random matrices
===============

.. currentmodule:: thewalrus.random

This submodule provides access to utility functions to generate random unitary, symplectic
and covariance matrices.

"""
import numpy as np
import scipy as sp

# ------------------------------------------------------------------------
# Random numbers and matrices                                           |
# ------------------------------------------------------------------------


def randnc(*arg):
    """Normally distributed array of random complex numbers."""
    return np.random.randn(*arg) + 1j * np.random.randn(*arg)


def random_covariance(N, hbar=2, pure=False, block_diag=False):
    r"""Random covariance matrix.

    Args:
        N (int): number of modes
        hbar (float): the value of :math:`\hbar` to use in the definition
            of the quadrature operators :math:`x` and :math:`p`
        pure (bool): If True, a random covariance matrix corresponding
            to a pure state is returned.
        block_diag (bool): If True, uses passive Gaussian transformations that are orthogonal
            instead of unitary. This implies that the positions :math:`x` do not mix with
            the momenta :math:`p` and thus the covariance matrix is block diagonal.

    Returns:
        array: random :math:`2N\times 2N` covariance matrix
    """
    S = random_symplectic(N, block_diag=block_diag)

    if pure:
        return (hbar / 2) * S @ S.T

    nbar = 2 * np.abs(np.random.random(N)) + 1
    Vth = (hbar / 2) * np.diag(np.concatenate([nbar, nbar]))

    return S @ Vth @ S.T


def random_symplectic(N, passive=False, block_diag=False, scale=1.0):
    r"""Random symplectic matrix representing a Gaussian transformation.

    The squeezing parameters :math:`r` for active transformations are randomly
    sampled from the standard normal distribution, while passive transformations
    are randomly sampled from the Haar measure. Note that for the Symplectic
    group there is no notion of Haar measure since this is group is not compact.

    Args:
        N (int): number of modes
        passive (bool): If True, returns a passive Gaussian transformation (i.e.,
            one that preserves photon number). If False, returns an active
            transformation.
        block_diag (bool): If True, uses passive Gaussian transformations that are orthogonal
            instead of unitary. This implies that the positions :math:`q` do not mix with
            the momenta :math:`p` and thus the symplectic operator is block diagonal
        scale (float): Sets the scale of the random values used as squeezing parameters.
            They will range from 0 to :math:`\sqrt{2}\texttt{scale}`

    Returns:
        array: random :math:`2N\times 2N` symplectic matrix
    """
    U = random_interferometer(N, real=block_diag)
    O = np.block([[U.real, -U.imag], [U.imag, U.real]])

    if passive:
        return O

    U = random_interferometer(N, real=block_diag)
    P = np.block([[U.real, -U.imag], [U.imag, U.real]])

    r = scale * np.abs(randnc(N))
    Sq = np.diag(np.concatenate([np.exp(-r), np.exp(r)]))

    return O @ Sq @ P


def random_interferometer(N, real=False):
    r"""Random unitary matrix representing an interferometer.
    For more details, see :cite:`mezzadri2006`.

    Args:
        N (int): number of modes
        real (bool): return a random real orthogonal matrix

    Returns:
        array: random :math:`N\times N` unitary distributed with the Haar measure
    """
    if real:
        z = np.random.randn(N, N)
    else:
        z = randnc(N, N) / np.sqrt(2.0)
    q, r = sp.linalg.qr(z)
    d = np.diagonal(r)
    ph = d / np.abs(d)
    U = np.multiply(q, ph, q)
    return U


def random_block_interferometer(N, top_one=True, real=False):
    r"""Generates a random interferometer with blocks of at most size 2.

    Args:
        N (int): number of modes
        top_one (bool): if True places a `1\times1` interferometer in the top-left most block
        real (bool): return a random real orthogonal matrix

    Returns:
        array: random :math:`N\times N` unitary with the specified block structure
    """
    if N % 2 == 0:
        if top_one:
            u2s = [random_interferometer(2, real=real) for i in range(((N // 2) - 1))]
            u0 = random_interferometer(1, real=real)
            u1 = random_interferometer(1, real=real)
            return sp.linalg.block_diag(u0, *u2s, u1)

        u2s = [random_interferometer(2, real=real) for i in range(N // 2)]
        return sp.linalg.block_diag(*u2s)

    u2s = [random_interferometer(2, real=real) for i in range((N - 1) // 2)]
    u0 = random_interferometer(1, real=real)
    if top_one:
        return sp.linalg.block_diag(u0, *u2s)
    return sp.linalg.block_diag(*u2s, u0)


def random_banded_interferometer(N, w, top_one_init=True, real=False):
    r"""Generates a banded unitary matrix.

    Args:
        N (int): number of modes
        w (int): bandwidth
        top_one_init (bool): if True places a `1\times1` interferometer in the top-left-most block of the first matrix in the product
        real (bool): return a random real orthogonal matrix

    Returns:
        array: random :math:`N\times N` unitary with the specified block structure
    """
    if N < w + 1:
        raise ValueError("The bandwidth can be at most one minus the size of the matrix.")
    if N == w + 1:
        return random_interferometer(N, real=real)
    U = sp.linalg.block_diag(*[random_interferometer(1, real=real) for _ in range(N)])
    for _ in range(w):
        U = U @ random_block_interferometer(N, top_one=top_one_init, real=real)
        top_one_init = not top_one_init
    return U



# Copyright 2019-2020 Xanadu Quantum Technologies Inc.

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
Functions for calculating the photon number distributions of various states.
"""

import numpy as np
from scipy.stats import nbinom
from scipy.special import binom, hyp2f1

from .conversions import Amat
from .gaussian_checks import is_pure_cov


def pure_state_distribution(cov, cutoff=50, hbar=2, padding_factor=2):
    r"""Calculates the total photon number distribution of a pure state
    with zero mean.

    Args:
        cov (array): :math:`2N\times 2N` covariance matrix in xp-ordering
        cutoff (int): Fock cutoff
        tol (float): tolerance for determining if displacement is negligible
        hbar (float): the value of :math:`\hbar` in the commutation
        padding_factor (int): expanded size of the photon distribution to avoid accumulation of errors

    Returns:
        (array): Total photon number distribution
    """
    if is_pure_cov(cov):
        A = Amat(cov, hbar=hbar)
        (n, _) = A.shape
        N = n // 2
        B = A[0:N, 0:N]
        rs = np.arctanh(np.linalg.svd(B, compute_uv=False))
        return _convolve_squeezed_state_distribution(
            rs, cutoff=cutoff, padding_factor=padding_factor
        )[0:cutoff]
    raise ValueError("The Gaussian state is not pure")


def _squeezed_state_distribution(s, cutoff=50, N=1):
    """Generate the photon number distribution of :math:`N` identical single mode squeezed states.

    Args:
        s (float): squeezing parameter
        cutoff (int): Fock cutoff
        N (float): number of squeezed states

    Returns:
        (array): Photon number distribution
    """
    r = 0.5 * N
    q = 1.0 - np.tanh(s) ** 2
    N = cutoff // 2
    ps_tot = np.zeros(cutoff)
    if cutoff % 2 == 0:
        ps = nbinom.pmf(np.arange(N), p=q, n=r)
        ps_tot[0::2] = ps
    else:
        ps = nbinom.pmf(np.arange(N + 1), p=q, n=r)
        ps_tot[0:-1][0::2] = ps[0:-1]
        ps_tot[-1] = ps[-1]

    return ps_tot


def _convolve_squeezed_state_distribution(s, cutoff=50, padding_factor=2):
    """Generates the total photon number distribution of single mode squeezed states with different squeezing values.

    Args:
        s (array): array of squeezing parameters
        cutoff (int): Fock cutoff
        padding_factor (int): expanded size of the photon distribution to avoid accumulation of errors
    Returns:
        (array[int]): total photon number distribution
    """
    scale = padding_factor
    cutoff_sc = scale * cutoff
    ps = np.zeros(cutoff_sc)
    ps[0] = 1.0
    for s_val in s:
        ps = np.convolve(ps, _squeezed_state_distribution(s_val, cutoff_sc))[0:cutoff_sc]
    return ps


def total_photon_number_distribution(n, k, s, eta, pref=1.0):
    r"""Probability of observing a total of :math:`n` photons when :math:`k` identical
    single-mode squeezed vacua with squeezing parameter :math:`s` undergo loss by transmission :math:`\eta`.

    For the derivation see Appendix E of `'Quantum Computational Supremacy via High-Dimensional Gaussian Boson Sampling',
    Deshpande et al. <https://arxiv.org/abs/2102.12474>`_.


    Args:
        n (int): number of photons
        k (int): number of squeezed modes
        s (float): squeezing parameter
        eta (float): transmission parameter, between 0 and 1 inclusive
        pref (float): use to return the probability times ``pref**n``
    Returns:
        (float): probability of observing a total of ``n`` photons or the probability times ``pref ** n``.
    """
    if n % 2 == 0:
        peven = (
            binom(-1 + k / 2.0 + n / 2.0, n / 2.0)
            * hyp2f1(0.5 + n / 2.0, k / 2.0 + n / 2.0, 0.5, (1 - eta) ** 2 * np.tanh(s) ** 2)
            * (1 / np.cosh(s)) ** k
            * (pref * eta * np.tanh(s)) ** n
        )
        return peven

    podd = (
        (1 + n)
        * (1 - eta)
        * binom((-1 + k + n) / 2.0, (1 + n) / 2.0)
        * hyp2f1((2 + n) / 2.0, (1 + k + n) / 2.0, 1.5, (1 - eta) ** 2 * np.tanh(s) ** 2)
        * (1 / np.cosh(s)) ** k
        * np.tanh(s)
        * (pref * eta * np.tanh(s)) ** n
    )
    return podd


def characteristic_function(
    k, s, eta, mu, max_iter=10000, delta=1e-14, poly_corr=None
):  # pylint: disable=too-many-arguments
    r"""Calculates the expectation value of the characteristic function
    :math:`\langle n^m \exp(mu n) \rangle` where :math:`n` is the total photon number of :math:`k` identical
    single-mode squeezed vacua with squeezing parameter :math:`s` undergoing loss by
    transmission :math:`\eta`.

    Args:
        k (int): number of squeezed modes
        s (float): squeezing parameter
        eta (float): transmission parameter, between 0 and 1 inclusive
        mu (float): value at which to evaluate the characteristic function
        max_iter (int): maximum number of terms allowed in the sum
        delta (float): fractional change in the sum after which the sum is stopped
        poly_corr (int): give the value of the exponent :math:`m` of the polynomial correction

    Returns:
        (float): the expected value of the moment generation function
    """

    if poly_corr is None or poly_corr == 0:
        f = lambda x: 1
    else:
        f = lambda x: x ** poly_corr

    if s == 0 or eta == 0:
        return f(0)

    pref = np.exp(mu)
    tot_sum = f(0) * total_photon_number_distribution(0, k, s, eta, pref=pref)
    converged = False

    i = 1
    prev_addend = tot_sum
    while converged is False and i < max_iter:
        old_tot_sum = tot_sum
        addend = f(i) * total_photon_number_distribution(i, k, s, eta, pref=pref)
        tot_sum += addend
        i += 1
        # Note that we check that the sum of the last *two* values does not change the net
        # sum much, this is because for eta=0 the distrobution does not have support over
        # the odd integers.
        ratio = (addend + prev_addend) / old_tot_sum
        if ratio < delta:
            converged = True
        prev_addend = addend
    return tot_sum



# Copyright 2019 Xanadu Quantum Technologies Inc.

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
Torontonian Python interface
"""
import numpy as np
import numba
from thewalrus.quantum.conversions import Xmat, Amat
from ._hafnian import reduction


def tor(A):
    """Returns the Torontonian of a matrix.

    Args:
        A (array): a square array of even dimensions.

    Returns:
        np.float64 or np.complex128: the torontonian of matrix A.
    """
    if not isinstance(A, np.ndarray):
        raise TypeError("Input matrix must be a NumPy array.")

    matshape = A.shape

    if matshape[0] != matshape[1]:
        raise ValueError("Input matrix must be square.")

    return numba_tor(A)


@numba.jit(nopython=True)
def combinations(pool, r):  # pragma: no cover
    """Numba implementation of `itertools.combinations`.
    As itertools.combinations not callable from numba decorated functions.

    Taken from: https://stackoverflow.com/a/61393666

    Args:
        pool (Sequence): sequence to take combinations from

        r (int) : number of elements of combination
    Yields:
        results (list) : length r combination from pool
    """
    n = len(pool)
    indices = list(range(r))
    empty = not (n and (0 < r <= n))

    if not empty:
        result = [pool[i] for i in indices]
        yield result

    while not empty:
        i = r - 1
        while i >= 0 and indices[i] == i + n - r:
            i -= 1
        if i < 0:
            empty = True
        else:
            indices[i] += 1
            for j in range(i + 1, r):
                indices[j] = indices[j - 1] + 1

            result = [pool[i] for i in indices]
            yield result


@numba.jit(nopython=True)
def powerset(parent_set):  # pragma: no cover
    """Generates the powerset, the set of all the subsets, of its input. Does not include the empty set.

    Args:
        parent_set (Sequence) : sequence to take powerset from
    Return:
        subset (tuple) : subset of parent_set
    """
    n = len(parent_set)
    for i in range(n + 1):
        for subset in combinations(parent_set, i):
            yield subset


@numba.jit(nopython=True)
def nb_block(X):  # pragma: no cover
    """Numba implementation of `np.block`.
    Only suitable for 2x2 blocks.

    Taken from: https://stackoverflow.com/a/57562911

    Args:
        X (array) : arrays for blocks of matrix
    Return:
        array : the block matrix from X
    """
    xtmp1 = np.concatenate(X[0], axis=1)
    xtmp2 = np.concatenate(X[1], axis=1)
    return np.concatenate((xtmp1, xtmp2), axis=0)


@numba.jit(nopython=True)
def numba_ix(arr, rows, cols):  # pragma: no cover
    """Numba implementation of `np.ix_`.
    Required due to numba lacking support for advanced numpy indexing.

    Args:
        arr (array) : matrix to take submatrix of
        rows (array) : rows to be selected in submatrix
        cols (array) : columns to be selected in submatrix

    Return:
        array: selected submatrix of arr, of shape `(len(rows), len(cols))`

    """
    return arr[rows][:, cols]


@numba.jit(nopython=True)
def Qmat_numba(cov, hbar=2):  # pragma: no cover
    r"""Numba compatible version of `thewalrus.quantum.Qmat`

    Returns the :math:`Q` Husimi matrix of the Gaussian state.
    Args:
        cov (array): :math:`2N\times 2N xp-` Wigner covariance matrix
        hbar (float): the value of :math:`\hbar` in the commutation
            relation :math:`[\x,\p]=i\hbar`.
    Returns:
        array: the :math:`Q` matrix.
    """
    # number of modes
    N = len(cov) // 2
    I = np.identity(N)

    x = cov[:N, :N] * (2.0 / hbar)
    xp = cov[:N, N:] * (2.0 / hbar)
    p = cov[N:, N:] * (2.0 / hbar)
    # the (Hermitian) matrix elements <a_i^\dagger a_j>
    aidaj = (x + p + 1j * (xp - xp.T) - 2 * I) / 4
    # the (symmetric) matrix elements <a_i a_j>
    aiaj = (x - p + 1j * (xp + xp.T)) / 4

    # calculate the covariance matrix sigma_Q appearing in the Q function:
    Q = nb_block(((aidaj, aiaj.conj()), (aiaj, aidaj.conj()))) + np.identity(2 * N)
    return Q


@numba.jit(nopython=True)
def threshold_detection_prob_displacement(mu, cov, det_pattern, hbar=2):  # pragma: no cover
    r"""Threshold detection probabilities for Gaussian states with displacement.
    Formula from Jake Bulmer and Stefano Paesani.


    Args:
        mu (1d array) : means of xp Gaussian Wigner function
        cov (2d array) : : xp Wigner covariance matrix
        det_pattern (1d numpy array) : array of {0,1} to describe the threshold detection outcome
        hbar (float): the value of :math:`\hbar` in the commutation relation :math:`[\x,\p]=i\hbar`.

    Returns:
        np.float64 : probability of detection pattern
    """
    det_pattern = np.asarray(det_pattern).astype(np.int8)

    m = len(cov)
    assert cov.shape == (m, m)
    assert m % 2 == 0
    n = m // 2

    means_x = mu[:n]
    means_p = mu[n:]
    avec = np.concatenate((means_x + 1j * means_p, means_x - 1j * means_p), axis=0) / np.sqrt(
        2 * hbar
    )

    Q = Qmat_numba(cov, hbar=hbar)

    if max(det_pattern) > 1:
        raise ValueError(
            "When using threshold detectors, the detection pattern can contain only 1s or 0s."
        )

    nonzero_idxs = np.where(det_pattern == 1)[0]
    zero_idxs = np.where(det_pattern == 0)[0]

    ii1 = np.concatenate((nonzero_idxs, nonzero_idxs + n), axis=0)
    ii0 = np.concatenate((zero_idxs, zero_idxs + n), axis=0)

    Qaa = numba_ix(Q, ii0, ii0)
    Qab = numba_ix(Q, ii0, ii1)
    Qba = numba_ix(Q, ii1, ii0)
    Qbb = numba_ix(Q, ii1, ii1)

    Qaa_inv = np.linalg.inv(Qaa)
    Qcond = Qbb - Qba @ Qaa_inv @ Qab

    avec_a = avec[ii0]
    avec_b = avec[ii1]
    avec_cond = avec_b - Qba @ Qaa_inv @ avec_a

    p0a_fact_exp = np.exp(avec_a @ Qaa_inv @ avec_a.conj() * (-0.5)).real
    p0a_fact_det = np.sqrt(np.linalg.det(Qaa).real)
    p0a = p0a_fact_exp / p0a_fact_det

    n_det = len(nonzero_idxs)
    p_sum = 1.0  # empty set is not included in the powerset function so we start at 1
    for z in powerset(np.arange(n_det)):
        Z = np.asarray(z)
        ZZ = np.concatenate((Z, Z + n_det), axis=0)

        avec0 = avec_cond[ZZ]
        Q0 = numba_ix(Qcond, ZZ, ZZ)
        Q0inv = np.linalg.inv(Q0)

        fact_exp = np.exp(avec0 @ Q0inv @ avec0.conj() * (-0.5)).real
        fact_det = np.sqrt(np.linalg.det(Q0).real)

        p_sum += ((-1) ** len(Z)) * fact_exp / fact_det

    return p0a * p_sum


def threshold_detection_prob(
    mu, cov, det_pattern, hbar=2, atol=1e-10, rtol=1e-10
):  # pylint: disable=too-many-arguments
    r"""Threshold detection probabilities for Gaussian states.
    Formula from Jake Bulmer and Stefano Paesani.
    When state is displaced, threshold_detection_prob_displacement is called.
    Otherwise, tor is called.

    Args:
        mu (1d array) : means of xp Gaussian Wigner function
        cov (2d array) : : xp Wigner covariance matrix
        det_pattern (1d array) : array of {0,1} to describe the threshold detection outcome
        hbar (float): the value of :math:`\hbar` in the commutation relation :math:`[\x,\p]=i\hbar`.
        rtol (float): the relative tolerance parameter used in `np.allclose`
        atol (float): the absolute tolerance parameter used in `np.allclose`

    Returns:
        np.float64 : probability of detection pattern
    """

    if np.allclose(mu, 0, atol=atol, rtol=rtol):
        # no displacement
        n_modes = cov.shape[0] // 2
        Q = Qmat_numba(cov, hbar)
        O = Xmat(n_modes) @ Amat(cov, hbar=hbar)
        rpt2 = np.concatenate((det_pattern, det_pattern))
        Os = reduction(O, rpt2)
        return tor(Os) / np.sqrt(np.linalg.det(Q))
    det_pattern = np.asarray(det_pattern).astype(np.int8)
    return threshold_detection_prob_displacement(mu, cov, det_pattern, hbar)


@numba.jit(nopython=True)
def numba_tor(A):  # pragma: no cover
    """Returns the Torontonian of a matrix using numba.

    For more direct control, you may wish to call :func:`tor_real` or
    :func:`tor_complex` directly.

    The input matrix is cast to quadruple precision
    internally for a quadruple precision torontonian computation.

    Args:
        A (array): a square, symmetric array of even dimensions.

    Returns:
        np.float64 or np.complex128: the torontonian of matrix A.
    """
    n_det = A.shape[0] // 2
    p_sum = 1.0  # empty set is not included in the powerset function so we start at 1
    for z in powerset(np.arange(n_det)):
        Z = np.asarray(z)
        ZZ = np.concatenate((Z, Z + n_det), axis=0)
        A_ZZ = numba_ix(A, ZZ, ZZ)
        n = len(Z)
        p_sum += ((-1) ** n) / np.sqrt(np.linalg.det(np.eye(2 * n) - A_ZZ))

    return p_sum * (-1) ** (n_det)



# Copyright 2019-2020 Xanadu Quantum Technologies Inc.

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
Tests for various properties of covariance matrices as well as fidelity
calculations for Gaussian states.
"""
# pylint: disable=too-many-arguments

import numpy as np
from scipy.linalg import sqrtm

from ..symplectic import sympmat


def is_valid_cov(cov, hbar=2, rtol=1e-05, atol=1e-08):
    r"""Checks if the covariance matrix is a valid quantum covariance matrix.

    Args:
        cov (array): a covariance matrix
        hbar (float): value of hbar in the uncertainty relation
        rtol (float): the relative tolerance parameter used in `np.allclose`
        atol (float): the absolute tolerance parameter used in `np.allclose`

    Returns:
        (bool): whether the given covariance matrix is a valid covariance matrix
    """
    (n, m) = cov.shape
    if n != m:
        # raise ValueError("The input matrix must be square")
        return False
    if not np.allclose(cov, np.transpose(cov), rtol=rtol, atol=atol):
        # raise ValueError("The input matrix is not symmetric")
        return False
    if n % 2 != 0:
        # raise ValueError("The input matrix is of even dimension")
        return False

    nmodes = n // 2
    vals = np.linalg.eigvalsh(cov + 0.5j * hbar * sympmat(nmodes))
    vals[np.abs(vals) < atol] = 0.0
    if np.all(vals >= 0):
        # raise ValueError("The input matrix violates the uncertainty relation")
        return True

    return False


def is_pure_cov(cov, hbar=2, rtol=1e-05, atol=1e-08):
    r"""Checks if the covariance matrix is a valid quantum covariance matrix
    that corresponds to a quantum pure state

    Args:
        cov (array): a covariance matrix
        hbar (float): value of hbar in the uncertainty relation
        rtol (float): the relative tolerance parameter used in `np.allclose`
        atol (float): the absolute tolerance parameter used in `np.allclose`

    Returns:
        (bool): whether the given covariance matrix corresponds to a pure state
    """
    if is_valid_cov(cov, hbar=hbar, rtol=rtol, atol=atol):
        purity = 1 / np.sqrt(np.linalg.det(2 * cov / hbar))
        if np.allclose(purity, 1.0, rtol=rtol, atol=atol):
            return True

    return False


def is_classical_cov(cov, hbar=2, atol=1e-08):
    r"""Checks if the covariance matrix can be efficiently sampled.

    Args:
        cov (array): a covariance matrix
        hbar (float): value of hbar in the uncertainty relation
        atol (float): the absolute tolerance parameter used in `np.allclose`

    Returns:
        (bool): whether the given covariance matrix corresponds to a classical state
    """

    if is_valid_cov(cov, hbar=hbar, atol=atol):
        (n, _) = cov.shape
        vals = np.linalg.eigvalsh(cov - 0.5 * hbar * np.identity(n))
        vals[np.abs(vals) < atol] = 0.0

        if np.all(vals >= 0):
            return True
    return False


def fidelity(mu1, cov1, mu2, cov2, hbar=2, rtol=1e-05, atol=1e-08):
    r"""Calculates the fidelity between two Gaussian quantum states.
    For two pure states :math:`|\psi_1 \rangle,  \  |\psi_2 \rangle`
    the fidelity is given by :math:`|\langle \psi_1|\psi_2 \rangle|^2`

    Note that if the covariance matrices correspond to pure states this
    function reduces to the modulus square of the overlap of their state vectors.
    For the derivation see  `'Quantum Fidelity for Arbitrary Gaussian States', Banchi et al. <10.1103/PhysRevLett.115.260501>`_.

    The actual implementation used here corresponds to the *square* of Eq. 96 of
    `'Gaussian states and operations - a quick reference', Brask <https://arxiv.org/abs/2102.05748>`_.

    Args:
        mu1 (array): vector of means of the first state
        cov1 (array): covariance matrix of the first state
        mu2 (array): vector of means of the second state
        cov2 (array): covariance matrix of the second state
        hbar (float): value of hbar in the uncertainty relation
        rtol (float): the relative tolerance parameter used in `np.allclose`
        atol (float): the absolute tolerance parameter used in `np.allclose`

    Returns:
        (float): value of the fidelity between the two states
    """

    n0, n1 = cov1.shape
    m0, m1 = cov2.shape
    (l0,) = mu1.shape
    (l1,) = mu1.shape
    if not n0 == n1 == m0 == m1 == l0 == l1:
        raise ValueError("The inputs have incompatible shapes")

    # We first convert all the inputs to quantities where hbar = 1
    sigma1 = cov1 / hbar
    sigma2 = cov2 / hbar
    deltar = (mu1 - mu2) / np.sqrt(hbar)

    omega = sympmat(n0 // 2)  # The symplectic matrix

    sigma = sigma1 + sigma2
    sigma_inv = np.linalg.inv(sigma)
    vaux = omega.T @ sigma_inv @ (0.25 * omega + sigma2 @ omega @ sigma1)
    sqrtm_arg = np.identity(n0) + 0.25 * np.linalg.inv(vaux @ omega @ vaux @ omega)

    # The sqrtm function has issues with matrices that are close to zero, hence we branch
    if np.allclose(sqrtm_arg, 0, rtol=rtol, atol=atol):
        mat_sqrtm = np.zeros_like(sqrtm_arg)
    else:
        mat_sqrtm = sqrtm(sqrtm_arg)

    det_arg = 2 * (mat_sqrtm + np.identity(n0)) @ vaux
    f = np.sqrt(np.linalg.det(sigma_inv) * np.linalg.det(det_arg)) * np.exp(
        -0.5 * deltar @ sigma_inv @ deltar
    )
    # Note that we only take the square root and that we have a prefactor of 0.5
    # as opposed to 0.25 in Brask. This is because this function returns the square
    # of their fidelities.
    return f



# Copyright 2019 Xanadu Quantum Technologies Inc.

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
This module implements common shared matrix decompositions that are
used to perform gate decompositions.
"""
import numpy as np

from scipy.linalg import block_diag, sqrtm, schur
from thewalrus.symplectic import sympmat


def williamson(V, rtol=1e-05, atol=1e-08):
    r"""Williamson decomposition of positive-definite (real) symmetric matrix.

    See https://math.stackexchange.com/questions/1171842/finding-the-symplectic-matrix-in-williamsons-theorem/2682630#2682630
        and https://strawberryfields.ai/photonics/conventions/decompositions.html#williamson-decomposition
    Args:
        V (array[float]): positive definite symmetric (real) matrix
        rtol (float): the relative tolerance parameter used in ``np.allclose``
        atol (float): the absolute tolerance parameter used in ``np.allclose``

    Returns:
        tuple[array,array]: ``(Db, S)`` where ``Db`` is a diagonal matrix
            and ``S`` is a symplectic matrix such that :math:`V = S^T Db S`
    """
    (n, m) = V.shape

    if n != m:
        raise ValueError("The input matrix is not square")

    if not np.allclose(V, V.T, rtol=rtol, atol=atol):
        raise ValueError("The input matrix is not symmetric")

    if n % 2 != 0:
        raise ValueError("The input matrix must have an even number of rows/columns")

    n = n // 2
    omega = sympmat(n)
    vals = np.linalg.eigvalsh(V)

    for val in vals:
        if val <= 0:
            raise ValueError("Input matrix is not positive definite")

    Mm12 = sqrtm(np.linalg.inv(V)).real
    r1 = Mm12 @ omega @ Mm12
    s1, K = schur(r1)
    X = np.array([[0, 1], [1, 0]])
    I = np.identity(2)
    seq = []

    # In what follows I construct a permutation matrix p  so that the Schur matrix has
    # only positive elements above the diagonal
    # Also the Schur matrix uses the x_1,p_1, ..., x_n,p_n  ordering thus I permute using perm
    # to go to the ordering x_1, ..., x_n, p_1, ... , p_n

    for i in range(n):
        if s1[2 * i, 2 * i + 1] > 0:
            seq.append(I)
        else:
            seq.append(X)
    perm = np.array([2 * i for i in range(n)] + [2 * i + 1 for i in range(n)])
    p = block_diag(*seq)
    Kt = K @ p
    Ktt = Kt[:, perm]
    s1t = p @ s1 @ p
    dd = [1 / s1t[2 * i, 2 * i + 1] for i in range(n)]
    Db = np.diag(dd + dd)
    S = Mm12 @ Ktt @ sqrtm(Db)
    return Db, np.linalg.inv(S).T



# Copyright 2019 Xanadu Quantum Technologies Inc.

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
Fock gradients of Gaussian gates
================================

.. currentmodule:: thewalrus.fock_gradients

This module contains the Fock representation of the standard Gaussian gates as well as their gradients.

.. autosummary::
    :toctree: api

	displacement
	squeezing
	beamsplitter
	two_mode_squeezing
	mzgate
	grad_displacement
	grad_squeezing
	grad_beamsplitter
	grad_two_mode_squeezing
	grad_mzgate

"""
import numpy as np

from numba import jit


@jit(nopython=True)
def displacement(r, phi, cutoff, dtype=np.complex128):  # pragma: no cover
    r"""Calculates the matrix elements of the displacement gate using a recurrence relation.

    Args:
        r (float): displacement magnitude
        phi (float): displacement angle
        cutoff (int): Fock ladder cutoff
        dtype (data type): Specifies the data type used for the calculation

    Returns:
        array[complex]: matrix representing the displacement operation.
    """
    D = np.zeros((cutoff, cutoff), dtype=dtype)
    sqrt = np.sqrt(np.arange(cutoff, dtype=dtype))
    mu = np.array([r * np.exp(1j * phi), -r * np.exp(-1j * phi)])

    D[0, 0] = np.exp(-0.5 * r ** 2)
    for m in range(1, cutoff):
        D[m, 0] = mu[0] / sqrt[m] * D[m - 1, 0]

    for m in range(cutoff):
        for n in range(1, cutoff):
            D[m, n] = mu[1] / sqrt[n] * D[m, n - 1] + sqrt[m] / sqrt[n] * D[m - 1, n - 1]

    return D


@jit(nopython=True)
def grad_displacement(T, r, phi):  # pragma: no cover
    r"""Calculates the gradients of the displacement gate with respect to the displacement magnitude and angle.

    Args:
        T (array[complex]): array representing the gate
        r (float): displacement magnitude
        phi (float): displacement angle

    Returns:
        tuple[array[complex], array[complex]]: The gradient of the displacement gate with respect to r and phi
    """
    cutoff = T.shape[0]
    dtype = T.dtype
    ei = np.exp(1j * phi)
    eic = np.exp(-1j * phi)
    alpha = r * ei
    alphac = r * eic
    sqrt = np.sqrt(np.arange(cutoff, dtype=dtype))
    grad_r = np.zeros((cutoff, cutoff), dtype=dtype)
    grad_phi = np.zeros((cutoff, cutoff), dtype=dtype)

    for m in range(cutoff):
        for n in range(cutoff):
            grad_r[m, n] = -r * T[m, n] + sqrt[m] * ei * T[m - 1, n] - sqrt[n] * eic * T[m, n - 1]
            grad_phi[m, n] = (
                sqrt[m] * 1j * alpha * T[m - 1, n] + sqrt[n] * 1j * alphac * T[m, n - 1]
            )

    return grad_r, grad_phi


@jit(nopython=True)
def squeezing(r, theta, cutoff, dtype=np.complex128):  # pragma: no cover
    r"""Calculates the matrix elements of the squeezing gate using a recurrence relation.

    Args:
        r (float): squeezing magnitude
        theta (float): squeezing angle
        cutoff (int): Fock ladder cutoff
        dtype (data type): Specifies the data type used for the calculation

    Returns:
        array[complex]: matrix representing the squeezing gate.
    """
    S = np.zeros((cutoff, cutoff), dtype=dtype)
    sqrt = np.sqrt(np.arange(cutoff, dtype=dtype))

    eitheta_tanhr = np.exp(1j * theta) * np.tanh(r)
    sechr = 1.0 / np.cosh(r)
    R = np.array(
        [
            [-eitheta_tanhr, sechr],
            [sechr, np.conj(eitheta_tanhr)],
        ]
    )

    S[0, 0] = np.sqrt(sechr)
    for m in range(2, cutoff, 2):
        S[m, 0] = sqrt[m - 1] / sqrt[m] * R[0, 0] * S[m - 2, 0]

    for m in range(0, cutoff):
        for n in range(1, cutoff):
            if (m + n) % 2 == 0:
                S[m, n] = (
                    sqrt[n - 1] / sqrt[n] * R[1, 1] * S[m, n - 2]
                    + sqrt[m] / sqrt[n] * R[0, 1] * S[m - 1, n - 1]
                )
    return S


@jit(nopython=True)
def grad_squeezing(T, r, phi):  # pragma: no cover
    r"""Calculates the gradients of the squeezing gate with respect to the squeezing magnitude and angle

    Args:
        T (array[complex]): array representing the gate
        r (float): squeezing magnitude
        phi (float): squeezing angle

    Returns:
        tuple[array[complex], array[complex]]: The gradient of the squeezing gate with respect to the r and phi
    """
    cutoff = T.shape[0]
    dtype = T.dtype
    sqrt = np.sqrt(np.arange(cutoff, dtype=dtype))
    grad_r = np.zeros((cutoff, cutoff), dtype=dtype)
    grad_phi = np.zeros((cutoff, cutoff), dtype=dtype)

    sechr = 1.0 / np.cosh(r)
    tanhr = np.tanh(r)
    eiphi = np.exp(1j * phi)
    eiphiconj = np.exp(-1j * phi)

    for m in range(cutoff):
        for n in range(cutoff):
            grad_r[m, n] = (
                -0.5 * tanhr * T[m, n]
                - sechr * tanhr * sqrt[m] * sqrt[n] * T[m - 1, n - 1]
                - 0.5 * eiphi * sechr ** 2 * sqrt[m] * sqrt[m - 1] * T[m - 2, n]
                + 0.5 * eiphiconj * sechr ** 2 * sqrt[n] * sqrt[n - 1] * T[m, n - 2]
            )
            grad_phi[m, n] = (
                -0.5j * eiphi * tanhr * sqrt[m] * sqrt[m - 1] * T[m - 2, n]
                - 0.5j * eiphiconj * tanhr * sqrt[n] * sqrt[n - 1] * T[m, n - 2]
            )

    return grad_r, grad_phi


@jit(nopython=True)
def two_mode_squeezing(r, theta, cutoff, dtype=np.complex128):  # pragma: no cover
    """Calculates the matrix elements of the two-mode squeezing gate recursively.

    Args:
        r (float): squeezing magnitude
        theta (float): squeezing angle
        cutoff (int): Fock ladder cutoff
        dtype (data type): Specifies the data type used for the calculation

    Returns:
        array[float]: The Fock representation of the gate

    """
    sqrt = np.sqrt(np.arange(cutoff, dtype=dtype))
    Z = np.zeros((cutoff, cutoff, cutoff, cutoff), dtype=dtype)

    sc = 1.0 / np.cosh(r)
    eiptr = np.exp(-1j * theta) * np.tanh(r)
    R = -np.array(
        [
            [0, -np.conj(eiptr), -sc, 0],
            [-np.conj(eiptr), 0, 0, -sc],
            [-sc, 0, 0, eiptr],
            [0, -sc, eiptr, 0],
        ]
    )

    Z[0, 0, 0, 0] = sc

    # rank 2
    for n in range(1, cutoff):
        Z[n, n, 0, 0] = R[0, 1] * Z[n - 1, n - 1, 0, 0]

    # rank 3
    for m in range(cutoff):
        for n in range(m):
            p = m - n
            if 0 < p < cutoff:
                Z[m, n, p, 0] = R[0, 2] * sqrt[m] / sqrt[p] * Z[m - 1, n, p - 1, 0]

    # rank 4
    for m in range(cutoff):
        for n in range(cutoff):
            for p in range(cutoff):
                q = p - (m - n)
                if 0 < q < cutoff:
                    Z[m, n, p, q] = (
                        R[1, 3] * sqrt[n] / sqrt[q] * Z[m, n - 1, p, q - 1]
                        + R[2, 3] * sqrt[p] / sqrt[q] * Z[m, n, p - 1, q - 1]
                    )
    return Z


@jit(nopython=True)
def grad_two_mode_squeezing(T, r, theta):  # pragma: no cover
    """Calculates the gradients of the two-mode squeezing gate with respect to the squeezing magnitude and angle

    Args:
        T (array[complex]): array representing the gate
        r (float): squeezing magnitude
        theta (float): squeezing angle

    Returns:
        tuple[array[complex], array[complex]]: The gradient of the two-mode squeezing gate with respect to r and phi

    """
    cutoff = T.shape[0]
    dtype = T.dtype
    sechr = 1.0 / np.cosh(r)
    tanhr = np.tanh(r)
    ei = np.exp(1j * theta)
    eic = np.exp(-1j * theta)
    sqrt = np.sqrt(np.arange(cutoff, dtype=dtype))

    grad_r = np.zeros((cutoff, cutoff, cutoff, cutoff), dtype=dtype)
    grad_theta = np.zeros((cutoff, cutoff, cutoff, cutoff), dtype=dtype)

    grad_r[0, 0, 0, 0] = -sechr * tanhr

    # rank 2
    for n in range(1, cutoff):
        grad_r[n, n, 0, 0] = (
            -tanhr * T[n, n, 0, 0] + sqrt[n] * sqrt[n] * ei * sechr ** 2 * T[n - 1, n - 1, 0, 0]
        )
        grad_theta[n, n, 0, 0] = 1j * ei * tanhr * sqrt[n] * sqrt[n] * T[n - 1, n - 1, 0, 0]

    # rank 3
    for m in range(cutoff):
        for n in range(m):
            p = m - n
            if 0 < p < cutoff:
                grad_r[m, n, p, 0] = (
                    -tanhr * T[m, n, p, 0]
                    + sqrt[m] * sqrt[n] * ei * sechr ** 2 * T[m - 1, n - 1, p, 0]
                    - tanhr * sechr * sqrt[m] * sqrt[p] * T[m - 1, n, p - 1, 0]
                )
                grad_theta[m, n, p, 0] = 1j * ei * tanhr * sqrt[m] * sqrt[n] * T[m - 1, n - 1, p, 0]

    # rank 4
    for m in range(cutoff):
        for n in range(cutoff):
            for p in range(cutoff):
                for q in range(cutoff):
                    grad_r[m, n, p, q] = (
                        -tanhr * T[m, n, p, q]
                        + sqrt[m] * sqrt[n] * ei * sechr ** 2 * T[m - 1, n - 1, p, q]
                        - tanhr * sechr * sqrt[m] * sqrt[p] * T[m - 1, n, p - 1, q]
                        - tanhr * sechr * sqrt[n] * sqrt[q] * T[m, n - 1, p, q - 1]
                        - sqrt[p] * sqrt[q] * eic * sechr ** 2 * T[m, n, p - 1, q - 1]
                    )
                    grad_theta[m, n, p, q] = (
                        1j * ei * tanhr * sqrt[m] * sqrt[n] * T[m - 1, n - 1, p, q]
                        + 1j * eic * tanhr * sqrt[p] * sqrt[q] * T[m, n, p - 1, q - 1]
                    )

    return grad_r, grad_theta


@jit(nopython=True)
def beamsplitter(theta, phi, cutoff, dtype=np.complex128):  # pragma: no cover
    r"""Calculates the Fock representation of the beamsplitter.

    Args:
        theta (float): transmissivity angle of the beamsplitter. The transmissivity is :math:`t=\cos(\theta)`
        phi (float): reflection phase of the beamsplitter
        cutoff (int): Fock ladder cutoff
        dtype (data type): Specifies the data type used for the calculation

    Returns:
        array[float]: The Fock representation of the gate
    """
    sqrt = np.sqrt(np.arange(cutoff, dtype=dtype))
    ct = np.cos(theta)
    st = np.sin(theta) * np.exp(1j * phi)
    R = np.array(
        [
            [0, 0, ct, -np.conj(st)],
            [0, 0, st, ct],
            [ct, st, 0, 0],
            [-np.conj(st), ct, 0, 0],
        ]
    )

    Z = np.zeros((cutoff, cutoff, cutoff, cutoff), dtype=dtype)
    Z[0, 0, 0, 0] = 1.0

    # rank 3
    for m in range(cutoff):
        for n in range(cutoff - m):
            p = m + n
            if 0 < p < cutoff:
                Z[m, n, p, 0] = (
                    R[0, 2] * sqrt[m] / sqrt[p] * Z[m - 1, n, p - 1, 0]
                    + R[1, 2] * sqrt[n] / sqrt[p] * Z[m, n - 1, p - 1, 0]
                )

    # rank 4
    for m in range(cutoff):
        for n in range(cutoff):
            for p in range(cutoff):
                q = m + n - p
                if 0 < q < cutoff:
                    Z[m, n, p, q] = (
                        R[0, 3] * sqrt[m] / sqrt[q] * Z[m - 1, n, p, q - 1]
                        + R[1, 3] * sqrt[n] / sqrt[q] * Z[m, n - 1, p, q - 1]
                    )
    return Z


@jit(nopython=True)
def grad_beamsplitter(T, theta, phi):  # pragma: no cover
    r"""Calculates the gradients of the beamsplitter gate with respect to the transmissivity angle and reflection phase

    Args:
        T (array[complex]): array representing the gate
        theta (float): transmissivity angle of the beamsplitter. The transmissivity is :math:`t=\cos(\theta)`
        phi (float): reflection phase of the beamsplitter

    Returns:
        tuple[array[complex], array[complex]]: The gradient of the beamsplitter gate with respect to theta and phi
    """
    cutoff = T.shape[0]
    dtype = T.dtype
    sqrt = np.sqrt(np.arange(cutoff, dtype=dtype))
    grad_theta = np.zeros((cutoff, cutoff, cutoff, cutoff), dtype=dtype)
    grad_phi = np.zeros((cutoff, cutoff, cutoff, cutoff), dtype=dtype)

    ct = np.cos(theta)
    st = np.sin(theta)
    ei = np.exp(1j * phi)
    eic = np.exp(-1j * phi)

    # rank 3
    for m in range(cutoff):
        for n in range(cutoff - m):
            p = m + n
            if 0 < p < cutoff:
                grad_theta[m, n, p, 0] = (
                    -sqrt[m] * sqrt[p] * st * T[m - 1, n, p - 1, 0]
                    + sqrt[n] * sqrt[p] * ei * ct * T[m, n - 1, p - 1, 0]
                )
                grad_phi[m, n, p, 0] = 1j * sqrt[n] * sqrt[p] * ei * st * T[m, n - 1, p - 1, 0]

    for m in range(cutoff):
        for n in range(cutoff):
            for p in range(cutoff):
                q = m + n - p
                if 0 < q < cutoff:
                    grad_theta[m, n, p, q] = (
                        -sqrt[m] * sqrt[p] * st * T[m - 1, n, p - 1, q]
                        - sqrt[n] * sqrt[q] * st * T[m, n - 1, p, q - 1]
                        + sqrt[n] * sqrt[p] * ei * ct * T[m, n - 1, p - 1, q]
                        - sqrt[m] * sqrt[q] * eic * ct * T[m - 1, n, p, q - 1]
                    )
                    grad_phi[m, n, p, q] = (
                        1j * sqrt[n] * sqrt[p] * ei * st * T[m, n - 1, p - 1, q]
                        + 1j * sqrt[m] * sqrt[q] * eic * st * T[m - 1, n, p, q - 1]
                    )

    return grad_theta, grad_phi


@jit(nopython=True)
def mzgate(theta, phi, cutoff, dtype=np.complex128):  # pragma: no cover
    r"""Calculates the Fock representation of the Mach-Zehnder interferometer.

    Args:
        theta (float): internal phase of the Mach-Zehnder interferometer
        phi (float): external phase of the Mach-Zehnder interferometer
        cutoff (int): Fock ladder cutoff
        dtype (data type): Specifies the data type used for the calculation

    Returns:
        array[float]: The Fock representation of the gate
    """
    sqrt = np.sqrt(np.arange(cutoff, dtype=dtype))
    v = np.exp(1j * theta)
    u = np.exp(1j * phi)
    R = 0.5 * np.array(
        [
            [0, 0, u * (v - 1), 1j * (1 + v)],
            [0, 0, 1j * u * (1 + v), 1 - v],
            [u * (v - 1), 1j * u * (1 + v), 0, 0],
            [1j * (1 + v), 1 - v, 0, 0],
        ]
    )

    Z = np.zeros((cutoff, cutoff, cutoff, cutoff), dtype=dtype)
    Z[0, 0, 0, 0] = 1.0

    # rank 3
    for m in range(cutoff):
        for n in range(cutoff - m):
            p = m + n
            if 0 < p < cutoff:
                Z[m, n, p, 0] = (
                    R[0, 2] * sqrt[m] / sqrt[p] * Z[m - 1, n, p - 1, 0]
                    + R[1, 2] * sqrt[n] / sqrt[p] * Z[m, n - 1, p - 1, 0]
                )

    # rank 4
    for m in range(cutoff):
        for n in range(cutoff):
            for p in range(cutoff):
                q = m + n - p
                if 0 < q < cutoff:
                    Z[m, n, p, q] = (
                        R[0, 3] * sqrt[m] / sqrt[q] * Z[m - 1, n, p, q - 1]
                        + R[1, 3] * sqrt[n] / sqrt[q] * Z[m, n - 1, p, q - 1]
                    )
    return Z


@jit(nopython=True)
def grad_mzgate(T, theta, phi):  # pragma: no cover
    r"""Calculates the gradients of the Mach-Zehnder interferometer with respect to the transmissivity angle and reflection phase

    Args:
        T (array[complex]): array representing the gate
        theta (float): internal of the mzgate
        phi (float): external phase of the mzgate

    Returns:
        tuple[array[complex], array[complex]]: The gradient of the mzgate gate with respect to theta and phi
    """
    cutoff = T.shape[0]
    dtype = T.dtype
    sqrt = np.sqrt(np.arange(cutoff, dtype=dtype))
    grad_theta = np.zeros((cutoff, cutoff, cutoff, cutoff), dtype=dtype)
    grad_phi = np.zeros((cutoff, cutoff, cutoff, cutoff), dtype=dtype)

    v = np.exp(1j * theta)
    u = np.exp(1j * phi)

    # rank 3
    for m in range(cutoff):
        for n in range(cutoff - m):
            p = m + n
            if 0 < p < cutoff:
                grad_theta[m, n, p, 0] = (
                    1j * 0.5 * u * v * sqrt[m] * sqrt[p] * T[m - 1, n, p - 1, 0]
                    - 0.5 * u * v * sqrt[n] * sqrt[p] * T[m, n - 1, p - 1, 0]
                )
                grad_phi[m, n, p, 0] = (1j * 0.5 * u * v - 1j * 0.5 * u) * sqrt[m] * sqrt[p] * T[
                    m - 1, n, p - 1, 0
                ] - (0.5 * u + 0.5 * u * v) * sqrt[n] * sqrt[p] * T[m, n - 1, p - 1, 0]

    for m in range(cutoff):
        for n in range(cutoff):
            for p in range(cutoff):
                q = m + n - p
                if 0 < q < cutoff:
                    grad_theta[m, n, p, q] = (
                        1j * 0.5 * u * v * sqrt[m] * sqrt[p] * T[m - 1, n, p - 1, q]
                        - 0.5 * u * v * sqrt[n] * sqrt[p] * T[m, n - 1, p - 1, q]
                        - 0.5 * v * sqrt[m] * sqrt[q] * T[m - 1, n, p, q - 1]
                        - 1j * 0.5 * v * sqrt[n] * sqrt[q] * T[m, n - 1, p, q - 1]
                    )
                    grad_phi[m, n, p, q] = (1j * 0.5 * u * v - 1j * 0.5 * u) * sqrt[m] * sqrt[
                        p
                    ] * T[m - 1, n, p - 1, q] - (0.5 * u + 0.5 * u * v) * sqrt[n] * sqrt[p] * T[
                        m, n - 1, p - 1, q
                    ]

    return grad_theta, grad_phi



# Copyright 2019 Xanadu Quantum Technologies Inc.

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
Symplectic Operations
=====================

**Module name:** :mod:`thewalrus.symplectic`

.. currentmodule:: thewalrus.symplectic

Contains some Gaussian operations and auxiliary functions.

Auxiliary functions
-------------------

.. autosummary::
    expand
    expand_vector
    expand_passive
    reduced_state
    is_symplectic
    sympmat
    xxpp_to_xpxp
    xpxp_to_xxpp

Gaussian states
---------------

.. autosummary::
    vacuum_state


Gates and operations
--------------------

.. autosummary::
    two_mode_squeezing
    squeezing
    interferometer
    loss
    mean_photon_number
    beam_splitter
    rotation

Code details
^^^^^^^^^^^^
"""
import numpy as np


def expand(S, modes, N):
    r"""Expands a Symplectic matrix S to act on the entire subsystem.

    Args:
        S (array): a :math:`2M\times 2M` Symplectic matrix
        modes (Sequence[int]): the list of modes S acts on
        N (int): full size of the subsystem

    Returns:
        array: the resulting :math:`2N\times 2N` Symplectic matrix
    """
    M = len(S) // 2
    S2 = np.identity(2 * N, dtype=S.dtype)
    w = np.array(modes)

    S2[w.reshape(-1, 1), w.reshape(1, -1)] = S[:M, :M].copy()  # X
    S2[(w + N).reshape(-1, 1), (w + N).reshape(1, -1)] = S[M:, M:].copy()  # P
    S2[w.reshape(-1, 1), (w + N).reshape(1, -1)] = S[:M, M:].copy()  # XP
    S2[(w + N).reshape(-1, 1), w.reshape(1, -1)] = S[M:, :M].copy()  # PX

    return S2


def expand_vector(alpha, mode, N, hbar=2.0):
    """Returns the phase-space displacement vector associated to a displacement.

    Args:
        alpha (complex): complex displacement
        mode (int): mode index
        N (int): number of modes

    Returns:
        array: phase-space displacement vector of size 2*N
    """
    alpharealdtype = np.dtype(type(alpha))

    r = np.zeros(2 * N, dtype=alpharealdtype)
    r[mode] = np.sqrt(2 * hbar) * alpha.real
    r[N + mode] = np.sqrt(2 * hbar) * alpha.imag
    return r


def expand_passive(T, modes, N):
    r"""Returns the expanded linear optical transformation
    acting on specified modes, with identity acting on all other modes

    Args:
        T (array): square :math:`M \times M` matrix of linear optical transformation
        modes (array): the :math:`M` modes of the transformation
        N (int): number of modes in the new expanded transformation

    Returns:
        array: :math:`N \times N` array of expanded passive transformation
    """

    if T.shape[0] != T.shape[1]:
        raise ValueError("The input matrix is not square")

    if len(modes) != T.shape[0]:
        raise ValueError("length of modes must match the shape of T")

    T_expand = np.eye(N, dtype=T.dtype)
    T_expand[np.ix_(modes, modes)] = T

    return T_expand


def reduced_state(mu, cov, modes):
    r"""Returns the vector of means and the covariance matrix of the specified modes.

    Args:
        modes (int of Sequence[int]): indices of the requested modes

    Returns:
        tuple (means, cov): means is an array containing the vector of means,
        and cov is a square array containing the covariance matrix. Both use the :math:`xp`-ordering.
    """
    N = len(mu) // 2

    if modes == list(range(N)):
        # reduced state is full state
        return mu, cov

    # reduce state down to specified subsystems
    if isinstance(modes, int):
        modes = [modes]

    if np.any(np.array(modes) > N):
        raise ValueError("The specified modes cannot be larger than number of subsystems.")

    ind = np.concatenate([np.array(modes), np.array(modes) + N])
    rows = ind.reshape(-1, 1)
    cols = ind.reshape(1, -1)

    return mu.copy()[ind], cov.copy()[rows, cols]


def vacuum_state(modes, hbar=2.0, dtype=np.float64):
    r"""Returns the vacuum state.

    Args:
        modes (str): Returns the vector of means and the covariance matrix
        hbar (float): (default 2) the value of :math:`\hbar` in the commutation
            relation :math:`[\x,\p]=i\hbar`
        dtype (numpy.dtype): datatype to represent the covariance matrix and vector of means
    Returns:
        list[array]: the means vector and covariance matrix of the vacuum state
    """
    means = np.zeros((2 * modes), dtype=dtype)
    cov = np.identity(2 * modes, dtype=dtype) * hbar / 2
    state = [means, cov]
    return state


def squeezing(r, phi=None, dtype=np.float64):
    r"""Squeezing. In fock space this corresponds to:

    .. math::

         \exp(\tfrac{1}{2}r e^{i \phi} (a^2 - a^{\dagger 2}) ).


    By passing an array of squeezing parameters and phases, it applies a tensor product of squeezing operations.

    Args:
        r (Union[array, float]): squeezing magnitude
        phi (Union[array, float]): rotation parameter. If ``None``, then the function uses zeros of the same shape as ``r``.
        dtype (numpy.dtype): datatype to represent the Symplectic matrix.  Defaults to ``numpy.float64``.
    Returns:
        array: symplectic transformation matrix
    """
    # pylint: disable=assignment-from-no-return

    r = np.atleast_1d(r)

    if phi is None:
        phi = np.zeros_like(r)

    phi = np.atleast_1d(phi)

    M = len(r)
    S = np.identity(2 * M, dtype=dtype)

    for i, (r_i, phi_i) in enumerate(zip(r, phi)):
        S[i, i] = np.cosh(r_i) - np.sinh(r_i) * np.cos(phi_i)
        S[i, i + M] = -np.sinh(r_i) * np.sin(phi_i)
        S[i + M, i] = -np.sinh(r_i) * np.sin(phi_i)
        S[i + M, i + M] = np.cosh(r_i) + np.sinh(r_i) * np.cos(phi_i)

    return S


def two_mode_squeezing(r, phi, dtype=np.float64):
    """Two-mode squeezing.

    Args:
        r (float): squeezing magnitude
        phi (float): rotation parameter
        dtype (numpy.dtype): datatype to represent the Symplectic matrix

    Returns:
        array: symplectic transformation matrix
    """
    # pylint: disable=assignment-from-no-return
    cp = np.cos(phi, dtype=dtype)
    sp = np.sin(phi, dtype=dtype)
    ch = np.cosh(r, dtype=dtype)
    sh = np.sinh(r, dtype=dtype)

    S = np.array(
        [
            [ch, cp * sh, 0, sp * sh],
            [cp * sh, ch, sp * sh, 0],
            [0, sp * sh, ch, -cp * sh],
            [sp * sh, 0, -cp * sh, ch],
        ]
    )

    return S


def interferometer(U):
    """Interferometer.

    Args:
        U (array): unitary matrix

    Returns:
        array: symplectic transformation matrix
    """
    X = U.real
    Y = U.imag
    S = np.block([[X, -Y], [Y, X]])

    return S


def passive_transformation(mu, cov, T, hbar=2):
    r"""Perform a covariance matrix transformation for an arbitrary linear optical channel
    on an :math:`N` modes state mapping it to a to an :math:`M` modes state.

    Args:
        mu (array): :math:`2N`-length means vector
        cov (array): :math:`2N \times 2N` covariance matrix
        T (array): :math:`M \times N` linear optical transformation

    Keyword Args:
        hbar (float)=2: the value to use for hbar

    Returns:
        array: :math:`2M`-length transformed means vector
        array :math:`2M \times 2M` tranformed covariance matrix
    """

    P = interferometer(T)
    L = (hbar / 2) * (np.eye(P.shape[0]) - P @ P.T)

    cov = P @ cov @ P.T + L
    mu = P @ mu

    return mu, cov


# pylint: disable=too-many-arguments
def loss(mu, cov, T, mode, nbar=0, hbar=2):
    r"""Loss channel acting on a Gaussian state.

    Args:
        mu (array): means vector
        cov (array): covariance matri
        T (float): transmission; 1 corresponds to no
            loss, 0 to full loss.
        mode (int): mode to act on
        nbar (float): thermal mean population (default 0)
        hbar (float): (default 2) the value of :math:`\hbar` in the commutation
            relation :math:`[\x,\p]=i\hbar`

    Returns:
        tuple[array]: the means vector and covariance matrix of the resulting state
    """
    N = len(cov) // 2

    cov_res = cov.copy()
    mu_res = mu.copy()

    for m in (mode, mode + N):
        mu_res[m] *= np.sqrt(T)
        cov_res[m, :] *= np.sqrt(T)
        cov_res[:, m] *= np.sqrt(T)
        cov_res[m, m] += (1 - T) * (2 * nbar + 1) * hbar / 2

    return mu_res, cov_res


### Comment: This function strongly overlaps with `quantum.photon_number_mean`
### Wonder if it is worth removing it.
def mean_photon_number(mu, cov, hbar=2):
    r"""Calculates the mean photon number for a given one-mode state.

    Args:
        mu (array): length-2 vector of means
        cov (array): :math:`2\times 2` covariance matrix
        hbar (float): (default 2) the value of :math:`\hbar` in the commutation
            relation :math:`[\x,\p]=i\hbar`

    Returns:
        tuple: the photon number expectation and variance
    """
    ex = (np.trace(cov) + mu.T @ mu) / (2 * hbar) - 1 / 2
    var = (np.trace(cov @ cov) + 2 * mu.T @ cov @ mu) / (2 * hbar ** 2) - 1 / 4
    return ex, var


def beam_splitter(theta, phi, dtype=np.float64):
    """Beam-splitter.

    Args:
        theta (float): transmissivity parameter
        phi (float): phase parameter
        dtype (numpy.dtype): datatype to represent the Symplectic matrix

    Returns:
        array: symplectic-orthogonal transformation matrix of an interferometer with angles theta and phi
    """
    ct = np.cos(theta, dtype=dtype)
    st = np.sin(theta, dtype=dtype)
    eip = np.cos(phi, dtype=dtype) + 1j * np.sin(phi, dtype=dtype)
    U = np.array(
        [
            [ct, -eip.conj() * st],
            [eip * st, ct],
        ]
    )
    return interferometer(U)


def rotation(theta, dtype=np.float64):
    """Rotation gate.

    Args:
        theta (float): rotation angle
        dtype (numpy.dtype): datatype to represent the Symplectic matrix

    Returns:
        array: rotation matrix by angle theta
    """
    V = np.identity(1) * (np.cos(theta, dtype=dtype) + 1j * np.sin(theta, dtype=dtype))
    return interferometer(V)


def sympmat(N, dtype=np.float64):
    r"""Returns the matrix :math:`\Omega_n = \begin{bmatrix}0 & I_n\\ -I_n & 0\end{bmatrix}`

    Args:
        N (int): positive integer
        dtype (numpy.dtype): datatype to represent the Symplectic matrix

    Returns:
        array: :math:`2N\times 2N` array
    """
    I = np.identity(N, dtype=dtype)
    O = np.zeros_like(I, dtype=dtype)
    S = np.block([[O, I], [-I, O]])
    return S


def is_symplectic(S, rtol=1e-05, atol=1e-08):
    r"""Checks if matrix S is a symplectic matrix

    Args:
        S (array): a square matrix

    Returns:
        (bool): whether the given matrix is symplectic
    """
    n, m = S.shape
    if n != m:
        return False
    if n % 2 != 0:
        return False
    nmodes = n // 2
    Omega = sympmat(nmodes, dtype=S.dtype)

    return np.allclose(S.T @ Omega @ S, Omega, rtol=rtol, atol=atol)


def autonne(A, rtol=1e-05, atol=1e-08, svd_order=True):
    r"""Autonne-Takagi decomposition of a complex symmetric (not Hermitian!) matrix.

    Args:
        A (array): square, symmetric matrix
        rtol (float): the relative tolerance parameter between ``A`` and ``A.T``
        atol (float): the absolute tolerance parameter between ``A`` and ``A.T``
        svd_order (boolean): whether to return result by ordering the singular values of ``A`` in descending (``True``) or asceding (``False``) order.

    Returns:
        tuple[array, array]: (r, U), where r are the singular values,
        and U is the Autonne-Takagi unitary, such that :math:`A = U \diag(r) U^T`.
    """
    n, m = A.shape
    if n != m:
        raise ValueError("The input matrix is not square")
    if not np.allclose(A, A.T, rtol=rtol, atol=atol):
        raise ValueError("The input matrix is not symmetric")
    Areal = A.real
    Aimag = A.imag

    B = np.empty((2 * n, 2 * n))
    B[:n, :n] = Areal
    B[n : 2 * n, :n] = Aimag
    B[:n, n : 2 * n] = Aimag
    B[n : 2 * n, n : 2 * n] = -Areal
    vals, vects = np.linalg.eigh(B)
    U = vects[:n, n : 2 * n] + 1j * vects[n : 2 * n, n : 2 * n]
    if svd_order:
        return (vals[n : 2 * n])[::-1], U[:, ::-1]
    return vals[n : 2 * n], U


def xxpp_to_xpxp(S):
    """Permutes the entries of the input from xxpp ordering to xpxp ordering.

    Args:
        S (array): input even dimensional square matrix or array

    Returns:
        (array): permuted matrix or array
    """
    shape = S.shape
    n = shape[0]

    if n % 2 != 0:
        raise ValueError("The input array is not even-dimensional")

    n = n // 2
    ind = np.arange(2 * n).reshape(2, -1).T.flatten()

    if len(shape) == 2:
        if shape[0] != shape[1]:
            raise ValueError("The input matrix is not square")
        return S[:, ind][ind]

    return S[ind]


def xpxp_to_xxpp(S):
    """Permutes the entries of the input from xpxp ordering to xxpp ordering.

    Args:
        S (array): input even dimensional square matrix or vector

    Returns:
        (array): permuted matrix or vector
    """
    shape = S.shape
    n = shape[0]

    if n % 2 != 0:
        raise ValueError("The input array is not even-dimensional")

    n = n // 2
    ind = np.arange(2 * n).reshape(-1, 2).T.flatten()

    if len(shape) == 2:
        if shape[0] != shape[1]:
            raise ValueError("The input matrix is not square")
        return S[:, ind][ind]

    return S[ind]



# Copyright 2019-2020 Xanadu Quantum Technologies Inc.

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
Quantum algorithms
==================

.. currentmodule:: thewalrus.quantum

This submodule provides access to various utility functions that act on Gaussian
quantum states.

For more details on how the hafnian relates to various properties of Gaussian quantum
states, see:

* Kruse, R., Hamilton, C. S., Sansoni, L., Barkhofen, S., Silberhorn, C., & Jex, I.
  "Detailed study of Gaussian boson sampling." `Physical Review A 100, 032326 (2019)
  <https://journals.aps.org/pra/abstract/10.1103/PhysRevA.100.032326>`_

* Hamilton, C. S., Kruse, R., Sansoni, L., Barkhofen, S., Silberhorn, C., & Jex, I.
  "Gaussian boson sampling." `Physical Review Letters, 119(17), 170501 (2017)
  <https://journals.aps.org/prl/abstract/10.1103/PhysRevLett.119.170501>`_

* Quesada, N.
  "Franck-Condon factors by counting perfect matchings of graphs with loops."
  `Journal of Chemical Physics 150, 164113 (2019) <https://aip.scitation.org/doi/10.1063/1.5086387>`_

* Quesada, N., Helt, L. G., Izaac, J., Arrazola, J. M., Shahrokhshahi, R., Myers, C. R., & Sabapathy, K. K.
  "Simulating realistic non-Gaussian state preparation." `Physical Review A 100, 022341 (2019)
  <https://journals.aps.org/pra/abstract/10.1103/PhysRevA.100.022341>`_


Fock states and tensors
^^^^^^^^^^^^^^^^^^^^^^^

.. autosummary::

    pure_state_amplitude
    state_vector
    density_matrix_element
    density_matrix
    fock_tensor
    probabilities
    loss_mat
    update_probabilities_with_loss
    update_probabilities_with_noise
    find_classical_subsystem
    tvd_cutoff_bounds
    n_body_marginals

Adjacency matrices
^^^^^^^^^^^^^^^^^^

.. autosummary::

    adj_scaling
    adj_scaling_torontonian
    adj_to_qmat

Gaussian checks
^^^^^^^^^^^^^^^

.. autosummary::

    is_valid_cov
    is_pure_cov
    is_classical_cov
    fidelity

Conversions
^^^^^^^^^^^

.. autosummary::

    reduced_gaussian
    Xmat
    Qmat
    Covmat
    Amat
    complex_to_real_displacements
    real_to_complex_displacements

Means and variances
^^^^^^^^^^^^^^^^^^^

.. autosummary::

    photon_number_mean
    photon_number_mean_vector
    photon_number_covar
    photon_number_covmat
    photon_number_expectation
    photon_number_squared_expectation
    normal_ordered_expectation
    photon_number_moment
    s_ordered_expectation
    mean_clicks
    variance_clicks
    photon_number_cumulant
    click_cumulant

Photon number distributions
^^^^^^^^^^^^^^^^^^^^^^^^^^^

.. autosummary::

    pure_state_distribution
    total_photon_number_distribution
    characteristic_function


Details
^^^^^^^
"""
import warnings
import functools

from .fock_tensors import (
    pure_state_amplitude,
    state_vector,
    density_matrix_element,
    density_matrix,
    fock_tensor,
    probabilities,
    loss_mat,
    update_probabilities_with_loss,
    update_probabilities_with_noise,
    find_classical_subsystem,
    tvd_cutoff_bounds,
    n_body_marginals,
)

from .adjacency_matrices import (
    adj_scaling,
    adj_scaling_torontonian,
    adj_to_qmat,
)

from .gaussian_checks import (
    is_valid_cov,
    is_pure_cov,
    is_classical_cov,
    fidelity,
)

from .conversions import (
    reduced_gaussian,
    Xmat,
    Qmat,
    Covmat,
    Amat,
    complex_to_real_displacements,
    real_to_complex_displacements,
)

from .means_and_variances import (
    photon_number_mean,
    photon_number_mean_vector,
    photon_number_covar,
    photon_number_covmat,
    photon_number_expectation,
    photon_number_squared_expectation,
    photon_number_moment,
    photon_number_cumulant,
    normal_ordered_expectation,
    mean_clicks,
    variance_clicks,
    s_ordered_expectation,
    click_cumulant,
)

from .photon_number_distributions import (
    pure_state_distribution,
    total_photon_number_distribution,
    characteristic_function,
)

__all__ = [
    "pure_state_amplitude",
    "state_vector",
    "density_matrix_element",
    "density_matrix",
    "fock_tensor",
    "probabilities",
    "loss_mat",
    "update_probabilities_with_loss",
    "update_probabilities_with_noise",
    "find_classical_subsystem",
    "adj_scaling",
    "adj_scaling_torontonian",
    "adj_to_qmat",
    "is_valid_cov",
    "is_pure_cov",
    "is_classical_cov",
    "fidelity",
    "reduced_gaussian",
    "Xmat",
    "Qmat",
    "Covmat",
    "Amat",
    "complex_to_real_displacements",
    "real_to_complex_displacements",
    "photon_number_mean",
    "photon_number_mean_vector",
    "photon_number_covar",
    "photon_number_covmat",
    "photon_number_expectation",
    "photon_number_squared_expectation",
    "normal_ordered_expectation",
    "mean_clicks",
    "variance_clicks",
    "pure_state_distribution",
    "n_body_marginals",
    "tvd_cutoff_bounds",
    "photon_number_moment",
    "s_ordered_expectation",
    "photon_number_cumulant",
    "total_photon_number_distribution",
    "characteristic_function",
]


def deprecate(new_func):
    """Wrapper for deprecated functions to raise warning"""

    @functools.wraps(new_func)
    def wrapper(*args, **kwargs):
        warnings.warn(
            f"This function is deprecated and will be removed. Use {new_func.__name__} instead.",
            DeprecationWarning,
        )
        return new_func(*args, **kwargs)

    return wrapper


# old names for functions; remove in due time
Means = deprecate(real_to_complex_displacements)
Beta = deprecate(complex_to_real_displacements)
total_photon_num_dist_pure_state = deprecate(pure_state_distribution)
gen_Qmat_from_graph = deprecate(adj_to_qmat)
find_scaling_adjacency_matrix = deprecate(adj_scaling)
find_scaling_adjacency_matrix_torontonian = deprecate(adj_scaling_torontonian)
mean_number_of_clicks = deprecate(mean_clicks)
variance_number_of_clicks = deprecate(variance_clicks)



# Copyright 2021 Xanadu Quantum Technologies Inc.

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
r"""
Python library
==============

.. currentmodule:: thewalrus

This is the top level module of the The Walrus Python interface,
containing functions for computing the hafnian, loop hafnian,
and torontonian of matrices.

Algorithm terminology
---------------------

Eigenvalue hafnian algorithm
    The algorithm described in
    *A faster hafnian formula for complex matrices and its benchmarking on a supercomputer*,
    :cite:`bjorklund2018faster`.
    This algorithm scales like :math:`\mathcal{O}(n^3 2^{n/2})`, and supports calculation of
    the loop hafnian.

Recursive hafnian algorithm
    The algorithm described in *Counting perfect matchings as fast as Ryser* :cite:`bjorklund2012counting`.
    This algorithm scales like :math:`\mathcal{O}(n^4 2^{n/2})`. This algorithm does not
    currently support the loop hafnian.

Repeated hafnian algorithm
    The algorithm described in *From moments of sum to moments of product*, :cite:`kan2008moments`.
    This method is more efficient for matrices with repeated rows and columns, and supports calculation of
    the loop hafnian.

Approximate hafnian algorithm
    The algorithm described in *Polynomial time algorithms to approximate permanents and mixed discriminants
    within a simply exponential factor*, :cite:`barvinok1999polynomial`.
    This algorithm allows us to efficiently approximate the hafnian of
    matrices with non-negative elements. This is done by sampling determinants;
    the larger the number of samples taken, the higher the accuracy.

Batched hafnian algorithm
    An algorithm that allows the calculation of hafnians of all reductions of
    a given matrix up to the cutoff (resolution) provided. Internally, this algorithm
    makes use of the multidimensional Hermite polynomials as per
    *The calculation of multidimensional Hermite polynomials and Gram-Charlier coefficients*
    :cite:`berkowitz1970calculation`.

Low-rank hafnian algorithm
    An algorithm that allows to calculate the hafnian of an :math:`r`-rank matrix :math:`\bm{A}` of size :math:`n \times n`
    by factorizing it as :math:`\bm{A} = \bm{G} \bm{G}^T` where :math:`\bm{G}` is of size :math:`n \times r`. The algorithm
    is described in Appendix C of
    *A faster hafnian formula for complex matrices and its benchmarking on a supercomputer*,
    :cite:`bjorklund2018faster`.

Banded hafnian algorithm
    An algorithm that calculates the hafnian of a matrix :math:`\bm{A}` of size :math:`n \times n` with bandwidth
    :math:`w` by calculating and storing certain subhafnians dictated by the bandwidth. The algorithm
    is described in Section V of
    *Efficient sampling from shallow Gaussian quantum-optical circuits with local interactions*,
    :cite:`qi2020efficient`.

Sparse hafnian algorithm
    An algorithm that calculates the hafnian of a sparse matrix by taking advantage of the Laplace expansion and memoization, to store
    only the relevant paths that contribute non-zero values to the final calculation.



Python wrappers
---------------

.. autosummary::
    hafnian
    hafnian_repeated
    hafnian_batched
    tor
    perm
    permanent_repeated
    hermite_multidimensional
    hafnian_banded

Pure Python functions
---------------------

.. autosummary::
    reduction
    version
    low_rank_hafnian
"""
# pylint: disable=wrong-import-position
import os
import platform

import numpy as np

import thewalrus.quantum

from ._hafnian import (
    haf_complex,
    haf_int,
    haf_real,
    haf_rpt_complex,
    haf_rpt_real,
    hafnian,
    hafnian_repeated,
    reduction,
    hafnian_sparse,
    hafnian_banded,
)

from ._low_rank_haf import low_rank_hafnian
from ._hermite_multidimensional import (
    hafnian_batched,
    hermite_multidimensional,
    interferometer,
    grad_hermite_multidimensional,
)
from ._permanent import (
    perm,
    perm_complex,
    perm_real,
    permanent_repeated,
    perm_BBFG_real,
    perm_BBFG_complex,
)
from ._torontonian import (
    tor,
    threshold_detection_prob_displacement,
    threshold_detection_prob,
    numba_tor,
)
from ._version import __version__


__all__ = [
    "hafnian",
    "hafnian_repeated",
    "hafnian_batched",
    "hafnian_sparse",
    "hafnian_banded",
    "tor",
    "perm",
    "permanent_repeated",
    "reduction",
    "hermite_multidimensional",
    "grad_hermite_multidimensional",
    "version",
]


def version():
    r"""
    Get version number of The Walrus

    Returns:
      str: The package version number
    """
    return __version__


def about():
    """The Walrus information.

    Prints the installed version numbers for TW and its dependencies,
    and some system info. Please include this information in bug reports.

    **Example:**

    .. code-block:: pycon

        >>> tw.about()
        The Walrus: a Python library for for the calculation of hafnians, Hermite polynomials, and Gaussian boson sampling.
        Copyright 2018-2020 Xanadu Quantum Technologies Inc.

        Python version:            3.8.5
        Platform info:             Linux-5.8.18-1-MANJARO-x86_64-with-arch-Manjaro-Linux
        Installation path:         /home/username/xanadu/thewalrus/thewalrus
        The Walrus version:        0.17.0-dev
        Numpy version:             1.19.5
        Scipy version:             1.7.1
        SymPy version:             1.8
        Numba version:             0.53.1
        Cython version:            0.29.24
    """
    # pylint: disable=import-outside-toplevel
    import sys
    import numpy
    import scipy
    import sympy
    import numba

    # a QuTiP-style infobox
    print(
        "\nThe Walrus: a Python library for for the calculation of hafnians, Hermite polynomials, and Gaussian boson sampling."
    )
    print("Copyright 2018-2021 Xanadu Quantum Technologies Inc.\n")

    print("Python version:            {}.{}.{}".format(*sys.version_info[0:3]))
    print("Platform info:             {}".format(platform.platform()))
    print("Installation path:         {}".format(os.path.dirname(__file__)))
    print("The Walrus version:        {}".format(__version__))
    print("Numpy version:             {}".format(numpy.__version__))
    print("Scipy version:             {}".format(scipy.__version__))
    print("SymPy version:             {}".format(sympy.__version__))
    print("Numba version:             {}".format(numba.__version__))



# Copyright 2021 Xanadu Quantum Technologies Inc.

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
Hafnian Python interface
"""

from functools import lru_cache
from collections import Counter
from itertools import chain, combinations
from numba import jit

import numpy as np

from .libwalrus import haf_complex, haf_int, haf_real, haf_rpt_complex, haf_rpt_real


def input_validation(A, rtol=1e-05, atol=1e-08):
    """Checks that the matrix A satisfies the requirements for Hafnian calculation.

    These include:

    * That the ``A`` is a NumPy array
    * That ``A`` is square
    * That ``A`` does not contain any NaNs
    * That ``A`` is symmetric

    Args:
        A (array): a NumPy array.
        rtol (float): the relative tolerance parameter used in ``np.allclose``.
        atol (float): the absolute tolerance parameter used in ``np.allclose``.

    Returns:
        bool: returns True if the matrix satisfies all requirements.
    """

    if not isinstance(A, np.ndarray):
        raise TypeError("Input matrix must be a NumPy array.")

    n = A.shape

    if n[0] != n[1]:
        raise ValueError("Input matrix must be square.")

    if np.isnan(A).any():
        raise ValueError("Input matrix must not contain NaNs.")

    if not np.allclose(A, A.T, rtol=rtol, atol=atol):
        raise ValueError("Input matrix must be symmetric.")

    return True


def bandwidth(A):
    """Calculates the upper bandwidth of the matrix A.

    Args:
        A (array): input matrix

    Returns:
        (int): bandwidth of matrix
    """
    n, _ = A.shape
    for i in range(n - 1, 0, -1):
        vali = np.diag(A, i)
        if not np.allclose(vali, 0):
            return i
    return 0


def powerset(iterable):
    """Calculates the powerset of a list.

    Args:
        iterable (iterable): input list

    Returns:
        (chain): chain of all subsets of input list
    """
    return chain.from_iterable(combinations(iterable, r) for r in range(len(iterable) + 1))


def reduction(A, rpt):
    r"""Calculates the reduction of an array by a vector of indices.

    This is equivalent to repeating the ith row/column of :math:`A`, :math:`rpt_i` times.

    Args:
        A (array): matrix of size [N, N]
        rpt (Sequence): sequence of N positive integers indicating the corresponding rows/columns
            of A to be repeated.
    Returns:
        array: the reduction of A by the index vector rpt
    """
    rows = [i for sublist in [[idx] * j for idx, j in enumerate(rpt)] for i in sublist]

    if A.ndim == 1:
        return A[rows]

    return A[:, rows][rows]


# pylint: disable=too-many-arguments
def hafnian(
    A, loop=False, recursive=True, rtol=1e-05, atol=1e-08, quad=True, approx=False, num_samples=1000
):  # pylint: disable=too-many-arguments
    """Returns the hafnian of a matrix.

    For more direct control, you may wish to call :func:`haf_real`,
    :func:`haf_complex`, or :func:`haf_int` directly.

    Args:
        A (array): a square, symmetric array of even dimensions.
        loop (bool): If ``True``, the loop hafnian is returned. Default is ``False``.
        recursive (bool): If ``True``, the recursive algorithm is used. Note:
            the recursive algorithm does not currently support the loop hafnian.
            If ``loop=True``, then this keyword argument is ignored.
        rtol (float): the relative tolerance parameter used in ``np.allclose``.
        atol (float): the absolute tolerance parameter used in ``np.allclose``.
        quad (bool): If ``True``, the hafnian algorithm is performed with quadruple precision.
        approx (bool): If ``True``, an approximation algorithm is used to estimate the hafnian. Note that
            the approximation algorithm can only be applied to matrices ``A`` that only have non-negative entries.
        num_samples (int): If ``approx=True``, the approximation algorithm performs ``num_samples`` iterations
            for estimation of the hafnian of the non-negative matrix ``A``.

    Returns:
        np.int64 or np.float64 or np.complex128: the hafnian of matrix A.
    """
    # pylint: disable=too-many-return-statements,too-many-branches
    input_validation(A, rtol=rtol, atol=atol)

    matshape = A.shape

    if matshape == (0, 0):
        return 1

    if matshape[0] % 2 != 0 and not loop:
        return 0.0

    if np.allclose(np.diag(np.diag(A)), A, rtol=rtol, atol=atol):
        if loop:
            return np.prod(np.diag(A))
        return 0

    if matshape[0] % 2 != 0 and loop:
        A = np.pad(A, pad_width=((0, 1), (0, 1)), mode="constant")
        A[-1, -1] = 1.0

    matshape = A.shape

    if matshape[0] == 2:
        if loop:
            return A[0, 1] + A[0, 0] * A[1, 1]
        return A[0][1]

    if matshape[0] == 4:
        if loop:
            result = (
                A[0, 1] * A[2, 3]
                + A[0, 2] * A[1, 3]
                + A[0, 3] * A[1, 2]
                + A[0, 0] * A[1, 1] * A[2, 3]
                + A[0, 1] * A[2, 2] * A[3, 3]
                + A[0, 2] * A[1, 1] * A[3, 3]
                + A[0, 0] * A[2, 2] * A[1, 3]
                + A[0, 0] * A[3, 3] * A[1, 2]
                + A[0, 3] * A[1, 1] * A[2, 2]
                + A[0, 0] * A[1, 1] * A[2, 2] * A[3, 3]
            )
            return result

        return A[0, 1] * A[2, 3] + A[0, 2] * A[1, 3] + A[0, 3] * A[1, 2]

    if approx:
        if np.any(np.iscomplex(A)):
            raise ValueError("Input matrix must be real")

        if np.any(A < 0):
            raise ValueError("Input matrix must not have negative entries")

        return hafnian_approx(A, num_samples=num_samples)

    if A.dtype == complex:
        # array data is complex type
        if np.any(np.iscomplex(A)):
            # array values contain non-zero imaginary parts
            return haf_complex(A, loop=loop, recursive=recursive, quad=quad)

        # all array values have zero imaginary parts
        return haf_real(np.float64(A.real), loop=loop, recursive=recursive, quad=quad)

    if np.issubdtype(A.dtype, np.integer) and not loop:
        # array data is an integer type, and the user is not
        # requesting the loop hafnian
        return haf_int(np.int64(A))

    if np.issubdtype(A.dtype, np.integer) and loop:
        # array data is an integer type, and the user is
        # requesting the loop hafnian. Currently no
        # integer function for loop hafnians, have to instead
        # convert to float and use haf_real
        A = np.float64(A)

    return haf_real(
        A, loop=loop, recursive=recursive, quad=quad, approx=approx, nsamples=num_samples
    )


def hafnian_sparse(A, D=None, loop=False):
    r"""Returns the hafnian of a sparse symmetric matrix.

    This pure python implementation is very slow on full matrices, but faster the sparser a matrix is.
    As a rule of thumb, the crossover in runtime with respect to :func:`~.hafnian` happens around 50% sparsity.

    Args:
        A (array): the symmetric matrix of which we want to compute the hafnian
        D (set): set of indices that identify a submatrix. If ``None`` (default) it computes
            the hafnian of the whole matrix.
        loop (bool): If ``True``, the loop hafnian is returned. Default is ``False``.

    Returns:
        (float) hafnian of ``A`` or of the submatrix of ``A`` defined by the set of indices ``D``.
    """
    if D is None:
        D = frozenset(range(len(A)))
    else:
        D = frozenset(D)

    if not loop:
        A = A - np.diag(np.diag(A))

    if np.allclose(A, 0):
        return 0.0

    r, _ = np.nonzero(A)
    m = max(Counter(r).values())  # max nonzero values per row/column

    @lru_cache(maxsize=2 ** m)
    def indices(d, k):
        return d.intersection(set(np.nonzero(A[k])[0]))

    @lru_cache(maxsize=2 ** m)
    def lhaf(d: frozenset) -> float:
        if not d:
            return 1
        d_ = set(d)
        k = d_.pop()
        return sum(A[i, k] * lhaf(frozenset(d_).difference({i})) for i in indices(d, k))

    return lhaf(D)


def hafnian_repeated(A, rpt, mu=None, loop=False, rtol=1e-05, atol=1e-08):
    r"""Returns the hafnian of matrix with repeated rows/columns.

    The :func:`reduction` function may be used to show the resulting matrix
    with repeated rows and columns as per ``rpt``.

    As a result, the following are identical:

    >>> hafnian_repeated(A, rpt)
    >>> hafnian(reduction(A, rpt))

    However, using ``hafnian_repeated`` in the case where there are a large number
    of repeated rows and columns (:math:`\sum_{i}rpt_i \gg N`) can be
    significantly faster.

    .. note::

        If :math:`rpt=(1, 1, \dots, 1)`, then

        >>> hafnian_repeated(A, rpt) == hafnian(A)

    For more direct control, you may wish to call :func:`haf_rpt_real` or
    :func:`haf_rpt_complex` directly.

    Args:
        A (array): a square, symmetric :math:`N\times N` array.
        rpt (Sequence): a length-:math:`N` positive integer sequence, corresponding
            to the number of times each row/column of matrix :math:`A` is repeated.
        mu (array): a vector of length :math:`N` representing the vector of means/displacement.
            If not provided, ``mu`` is set to the diagonal of matrix ``A``. Note that this
            only affects the loop hafnian.
        loop (bool): If ``True``, the loop hafnian is returned. Default is ``False``.
        rtol (float): the relative tolerance parameter used in ``np.allclose``.
        atol (float): the absolute tolerance parameter used in ``np.allclose``.

    Returns:
        np.int64 or np.float64 or np.complex128: the hafnian of matrix A.
    """
    # pylint: disable=too-many-return-statements,too-many-branches
    input_validation(A, atol=atol, rtol=rtol)

    if len(rpt) != len(A):
        raise ValueError("the rpt argument must be 1-dimensional sequence of length len(A).")

    nud = np.array(rpt, dtype=np.int32)

    if not np.all(np.mod(nud, 1) == 0) or np.any(nud < 0):
        raise ValueError("the rpt argument must contain non-negative integers.")

    if np.all(nud == 0):
        return 1.0

    if np.sum(nud) % 2 != 0 and not loop:
        return 0.0

    if mu is None:
        mu = A.diagonal().copy()

    if np.allclose(A, 0, rtol=rtol, atol=atol):
        if loop:
            return np.prod(mu ** rpt)
        return 0

    if len(mu) != len(A):
        raise ValueError("Length of means vector must be the same length as the matrix A.")

    if complex in (A.dtype, mu.dtype):
        return haf_rpt_complex(A, nud, mu=mu, loop=loop)

    return haf_rpt_real(A, nud, mu=mu, loop=loop)


def hafnian_banded(A, loop=False, rtol=1e-05, atol=1e-08):
    """Returns the loop hafnian of a banded matrix.

    For the derivation see Section V of `'Efficient sampling from shallow Gaussian quantum-optical circuits with local interactions',
    Qi et al. <https://arxiv.org/abs/2009.11824>`_.

    Args:
        A (array): a square, symmetric array of even dimensions.

    Returns:
        np.int64 or np.float64 or np.complex128: the loop hafnian of matrix ``A``.
    """
    input_validation(A, atol=atol, rtol=rtol)
    (n, _) = A.shape
    w = bandwidth(A)
    if not loop:
        A = A - np.diag(np.diag(A))
    loop_haf = {(): 1, (1,): A[0, 0]}
    for t in range(1, n + 1):
        if t - 2 * w - 1 > 0:
            lower_end = set(range(1, t - 2 * w))
        else:
            lower_end = set()
        upper_end = set(range(1, t + 1))
        diff = [item for item in upper_end if item not in lower_end]
        # Makes sure set ordering is preserved when the difference of two sets is taken
        # This is also used in the if statement below
        ps = powerset(diff)
        lower_end = tuple(lower_end)
        for D in ps:
            if lower_end + D not in loop_haf:
                # pylint: disable=consider-using-generator
                loop_haf[lower_end + D] = sum(
                    [
                        A[i - 1, t - 1]
                        * loop_haf[
                            tuple([item for item in lower_end + D if item not in set((i, t))])
                        ]
                        for i in D
                    ]
                )

    return loop_haf[tuple(range(1, n + 1))]


@jit(nopython=True)
def _one_det(B):  # pragma: no cover
    """Calculates the determinant of an antisymmetric matrix with entries distributed
    according to a normal distribution, with scale equal to the entries of the symmetric matrix
    given as input.

    Args:
        B (array[float]): symmetric matrix

    Returns:
        float: determinant of the samples antisymmetric matrix
    """
    mat = np.empty_like(B, dtype=np.float64)
    n, m = B.shape
    for i in range(n):
        for j in range(m):
            mat[i, j] = B[i, j] * np.random.normal()
            mat[j, i] = -mat[i, j]
    return np.linalg.det(mat)


@jit(nopython=True)
def hafnian_approx(A, num_samples=1000):  # pragma: no cover
    """Returns the approximation to the hafnian of a matrix with non-negative entries.

    The approximation follows the stochastic Barvinok's approximation allowing the
    hafnian can be approximated as the sum of determinants of matrices.
    The accuracy of the approximation increases with increasing number of iterations.

    Args:
        B (array[float]): a symmetric matrix

    Returns:
        (float): approximate hafnian of the input
    """

    sqrtA = np.sqrt(A)
    return np.array([_one_det(sqrtA) for _ in range(num_samples)]).mean()



# Copyright 2019 Xanadu Quantum Technologies Inc.

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
Algorithms for hafnians of low-rank matrices.
"""

from functools import lru_cache
from itertools import product
import numpy as np
from sympy import symbols, expand
from scipy.special import factorial2


@lru_cache(maxsize=1000000)
def partitions(r, n):
    r"""Returns a list of lists with the r-partitions of the integer :math:`n`, i.e. the r-tuples of non-negative
    integers such that their sum is precisely :math:`n`.

    Note that there are :math:`n + r - 1 \choose r-1` such partitions.

    Args:
        r (int): number of partitions.
        n (int): integer to be partitioned.

    Returns:
        list: r-partitions of n.
    """
    if r == 1:
        return [[n]]

    new_combos = []
    for first_val in range(n + 1):
        rest = partitions(r - 1, n - first_val)
        new = [p[0] + p[1] for p in product([[first_val]], rest)]
        new_combos += new
    return new_combos


def low_rank_hafnian(G):
    r"""Returns the hafnian of the low rank matrix :math:`\bm{A} = \bm{G} \bm{G}^T` where :math:`\bm{G}` is rectangular of size
    :math:`n \times r`  with :math:`r \leq n`.

    Note that the rank of :math:`\bm{A}` is precisely :math:`r`.

    The hafnian is calculated using the algorithm described in Appendix C of
    *A faster hafnian formula for complex matrices and its benchmarking on a supercomputer*,
    :cite:`bjorklund2018faster`.

    Args:
        G (array): factorization of the low rank matrix A = G @ G.T.

    Returns:
        (complex): hafnian of A.
    """
    n, r = G.shape
    if n % 2 != 0:
        return 0
    if r == 1:
        return factorial2(n - 1) * np.prod(G)
    poly = 1
    x = symbols("x0:" + str(r))
    for k in range(n):
        term = 0
        for j in range(r):
            term += G[k, j] * x[j]
        poly = expand(poly * term)

    comb = partitions(r, n // 2)
    haf_val = 0.0
    for c in comb:
        monomial = 1
        facts = 1
        for i, pi in enumerate(c):
            monomial *= x[i] ** (2 * pi)
            facts = facts * factorial2(2 * pi - 1)
        haf_val += complex(poly.coeff(monomial) * facts)
    return haf_val



# Copyright 2019-2020 Xanadu Quantum Technologies Inc.

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
Functions for transforming one type of covariance-matrix-like object into
another as well as reducing a Gaussian state.
"""

import numpy as np


def reduced_gaussian(mu, cov, modes):
    r"""Returns the vector of means and the covariance matrix of the specified modes.

    Args:
        mu (array): a length-:math:`2N` ``np.float64`` vector of means.
        cov (array): a :math:`2N\times 2N` ``np.float64`` covariance matrix
            representing an :math:`N` mode quantum state.
        modes (int of Sequence[int]): indices of the requested modes

    Returns:
        tuple (means, cov): where means is an array containing the vector of means,
        and cov is a square array containing the covariance matrix.
    """
    N = len(mu) // 2

    # reduce rho down to specified subsystems
    if isinstance(modes, int):
        modes = [modes]

    if np.any(np.array(modes) > N):
        raise ValueError("Provided mode is larger than the number of subsystems.")

    if len(modes) == N:
        # reduced state is full state
        return mu, cov

    ind = np.concatenate([np.array(modes), np.array(modes) + N])
    rows = ind.reshape(-1, 1)
    cols = ind.reshape(1, -1)

    return mu[ind], cov[rows, cols]


def Xmat(N):
    r"""Returns the matrix :math:`X_n = \begin{bmatrix}0 & I_n\\ I_n & 0\end{bmatrix}`

    Args:
        N (int): positive integer

    Returns:
        array: :math:`2N\times 2N` array
    """
    I = np.identity(N)
    O = np.zeros_like(I)
    X = np.block([[O, I], [I, O]])
    return X


def Qmat(cov, hbar=2):
    r"""Returns the :math:`Q` Husimi matrix of the Gaussian state.

    Args:
        cov (array): :math:`2N\times 2N xp-` Wigner covariance matrix
        hbar (float): the value of :math:`\hbar` in the commutation
            relation :math:`[\x,\p]=i\hbar`.

    Returns:
        array: the :math:`Q` matrix.
    """
    # number of modes
    N = len(cov) // 2
    I = np.identity(N)

    x = cov[:N, :N] * 2 / hbar
    xp = cov[:N, N:] * 2 / hbar
    p = cov[N:, N:] * 2 / hbar
    # the (Hermitian) matrix elements <a_i^\dagger a_j>
    aidaj = (x + p + 1j * (xp - xp.T) - 2 * I) / 4
    # the (symmetric) matrix elements <a_i a_j>
    aiaj = (x - p + 1j * (xp + xp.T)) / 4

    # calculate the covariance matrix sigma_Q appearing in the Q function:
    # Q(alpha) = exp[-(alpha-beta).sigma_Q^{-1}.(alpha-beta)/2]/|sigma_Q|
    Q = np.block([[aidaj, aiaj.conj()], [aiaj, aidaj.conj()]]) + np.identity(2 * N)
    return Q


def Covmat(Q, hbar=2):
    r"""Returns the Wigner covariance matrix in the :math:`xp`-ordering of the Gaussian state.
    This is the inverse function of Qmat.

    Args:
        Q (array): :math:`2N\times 2N` Husimi Q matrix
        hbar (float): the value of :math:`\hbar` in the commutation
            relation :math:`[\x,\p]=i\hbar`.

    Returns:
        array: the :math:`xp`-ordered covariance matrix in the xp-ordering.
    """
    # number of modes
    n = len(Q) // 2
    I = np.identity(n)
    N = Q[0:n, 0:n] - I
    M = Q[n : 2 * n, 0:n]
    mm11a = 2 * (N.real + M.real) + np.identity(n)
    mm22a = 2 * (N.real - M.real) + np.identity(n)
    mm12a = 2 * (M.imag + N.imag)
    cov = np.block([[mm11a, mm12a], [mm12a.T, mm22a]])

    return (hbar / 2) * cov


def Amat(cov, hbar=2, cov_is_qmat=False):
    r"""Returns the :math:`A` matrix of the Gaussian state whose hafnian gives the photon number probabilities.

    Args:
        cov (array): :math:`2N\times 2N` covariance matrix
        hbar (float): the value of :math:`\hbar` in the commutation
            relation :math:`[\x,\p]=i\hbar`.
        cov_is_qmat (bool): if ``True``, it is assumed that ``cov`` is in fact the Q matrix.

    Returns:
        array: the :math:`A` matrix.
    """
    # number of modes
    N = len(cov) // 2
    X = Xmat(N)

    # inverse Q matrix
    if cov_is_qmat:
        Q = cov
    else:
        Q = Qmat(cov, hbar=hbar)

    Qinv = np.linalg.inv(Q)

    # calculate Hamilton's A matrix: A = X.(I-Q^{-1})*
    A = X @ (np.identity(2 * N) - Qinv).conj()
    return A


def complex_to_real_displacements(mu, hbar=2):
    r"""Returns the vector of complex displacements and conjugate displacements.

    Args:
        mu (array): length-:math:`2N` means vector
        hbar (float): the value of :math:`\hbar` in the commutation
            relation :math:`[\x,\p]=i\hbar`.

    Returns:
        array: the expectation values
        :math:`[\langle a_1\rangle, \langle a_2\rangle,\dots,\langle a_N\rangle, \langle a^\dagger_1\rangle, \dots, \langle a^\dagger_N\rangle]`
    """
    N = len(mu) // 2
    # mean displacement of each mode
    alpha = (mu[:N] + 1j * mu[N:]) / np.sqrt(2 * hbar)
    # the expectation values (<a_1>, <a_2>,...,<a_N>, <a^\dagger_1>, ..., <a^\dagger_N>)
    return np.concatenate([alpha, alpha.conj()])


def real_to_complex_displacements(beta, hbar=2):
    r"""Returns the vector of real quadrature displacements.

    Args:
        beta (array): length-:math:`2N` means bivector
        hbar (float): the value of :math:`\hbar` in the commutation
            relation :math:`[\x,\p]=i\hbar`.

    Returns:
        array: the quadrature expectation values
        :math:`[\langle q_1\rangle, \langle q_2\rangle,\dots,\langle q_N\rangle, \langle p_1\rangle, \dots, \langle p_N\rangle]`
    """

    N = len(beta) // 2
    alpha = beta[0:N]
    return np.sqrt(2 * hbar) * np.concatenate([alpha.real, alpha.imag])
